#!/usr/bin/env python

"""XCP info/exploration tool."""

import argparse
from pprint import pprint

from pyxcp.cmdline import ArgumentParser
from pyxcp.types import TryCommandResult


def getPagInfo(x):
    result = {}
    if x.slaveProperties.supportsCalpag:
        status, pag = x.try_command(x.getPagProcessorInfo)
        if status == TryCommandResult.OK:
            result["maxSegments"] = pag.maxSegments
            result["pagProperties"] = {}
            result["pagProperties"]["freezeSupported"] = pag.pagProperties.freezeSupported
            result["segments"] = []
            for i in range(pag.maxSegments):
                segment = {}
                status, std_info = x.try_command(x.getSegmentInfo, 0x01, i, 0, 0)
                if status == TryCommandResult.OK:
                    segment["maxPages"] = std_info.maxPages
                    segment["addressExtension"] = std_info.addressExtension
                    segment["maxMapping"] = std_info.maxMapping
                    segment["compressionMethod"] = std_info.compressionMethod
                    segment["encryptionMethod"] = std_info.encryptionMethod

                    status, seg_address = x.try_command(x.getSegmentInfo, 0x00, i, 0, 0)
                    status, seg_length = x.try_command(x.getSegmentInfo, 0x00, i, 1, 0)

                    segment["address"] = seg_address.basicInfo
                    segment["length"] = seg_length.basicInfo

                    result["segments"].append(segment)

                    status, pgi = x.try_command(x.getPageInfo, i, 0)
                    # print("PAGE:", pgi)
                    # for j in range(si.maxPages):
                    #    pgi = x.getPageInfo(i, j)
                    #    print(pgi)
                else:
                    break
    return result


def main():
    parser = argparse.ArgumentParser(description="XCP info/exploration tool.")
    parser.add_argument("--no-daq", action="store_true", help="Do not query DAQ information.")
    parser.add_argument("--no-pag", action="store_true", help="Do not query PAG information.")
    parser.add_argument("--no-pgm", action="store_true", help="Do not query PGM information.")
    parser.add_argument("--no-ids", action="store_true", help="Do not scan implemented IDs.")
    ap = ArgumentParser(parser)

    with ap.run() as x:
        x.connect()
        if x.slaveProperties.optionalCommMode:
            x.try_command(x.getCommModeInfo, extra_msg="availability signaled by CONNECT, this may be a slave configuration error.")
        print("\nSlave Properties:")
        print("=================")
        pprint(x.slaveProperties)

        if not ap.args.no_ids:
            result = x.id_scanner()
            print("\n")
            print("Implemented IDs:")
            print("================")
            for key, value in result.items():
                print(f"{key}: {value}", end="\n\n")

        cps = x.getCurrentProtectionStatus()
        print("\nProtection Status")
        print("=================")
        for k, v in cps.items():
            print(f"    {k:6s}: {v}")
        x.cond_unlock()

        if not ap.args.no_daq:
            print("\nDAQ Info:")
            print("=========")
            if x.slaveProperties.supportsDaq:
                daq_info = x.getDaqInfo()
                pprint(daq_info)

                daq_pro = daq_info["processor"]
                daq_properties = daq_pro["properties"]
                if x.slaveProperties.transport_layer == "CAN":
                    print("")
                    if daq_properties["pidOffSupported"]:
                        print("*** pidOffSupported -- i.e. one CAN-ID per DAQ-list.")
                    else:
                        print("*** NO support for PID_OFF")
                num_predefined = daq_pro["minDaq"]
                print("\nPredefined DAQ-Lists")
                print("====================")
                if num_predefined > 0:
                    print(f"There are {num_predefined} predefined DAQ-lists")
                    for idx in range(num_predefined):
                        print(f"DAQ-List #{idx}\n____________\n")
                        status, dm = x.try_command(x.getDaqListMode, idx)
                        if status == TryCommandResult.OK:
                            print(dm)
                        status, di = x.try_command(x.getDaqListInfo, idx)
                        if status == TryCommandResult.OK:
                            print(di)
                else:
                    print("*** NO Predefined DAQ-Lists")
            else:
                print("*** DAQ IS NOT SUPPORTED .")

        if not ap.args.no_pag:
            print("\nPAG Info:")
            print("=========")
            if x.slaveProperties.supportsCalpag:
                pgi = getPagInfo(x)
                pprint(pgi)
            else:
                print("*** PAGING IS NOT SUPPORTED.")

        if not ap.args.no_pgm:
            print("\nPGM Info:")
            print("=========")
            if x.slaveProperties.supportsPgm:
                status, pgm = x.try_command(x.getPgmProcessorInfo)
                if status == TryCommandResult.OK:
                    print(pgm)
            else:
                print("*** FLASH PROGRAMMING IS NOT SUPPORTED.")

        if x.slaveProperties.transport_layer == "CAN":
            print("\nTransport-Layer CAN:")
            print("====================")
            status, res = x.try_command(x.getSlaveID, 0)
            if status == TryCommandResult.OK:
                print("CAN identifier for CMD/STIM:\n", res)
            else:
                pass
                # print("*** GET_SLAVE_ID() IS NOT SUPPORTED.")  # no response from bc address ???

            print("\nPer DAQ-list Identifier")
            print("-----------------------")
            daq_id = 0
            while True:
                status, res = x.try_command(x.getDaqId, daq_id)
                if status == TryCommandResult.OK:
                    print(f"DAQ-list #{daq_id}:", res)
                    daq_id += 1
                else:
                    break
            if daq_id == 0:
                print("N/A")
        x.disconnect()
        print("\nDone.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Copy pyXCP examples to a given directory.
"""

import argparse
import sys
from pathlib import Path

from pyxcp import console


pyver = sys.version_info
if pyver.major == 3 and pyver.minor <= 9:
    import pkg_resources

    def copy_files_from_package(package_name: str, source_directory: str, args: argparse.Namespace) -> None:
        destination_directory = args.output_directory
        force = args.force
        for fn in pkg_resources.resource_listdir(package_name, source_directory):
            source_file = Path(pkg_resources.resource_filename(package_name, f"{source_directory}/{fn}"))
            if source_file.suffix == ".py":
                dest_file = Path(destination_directory) / fn
                if dest_file.exists() and not force:
                    console.print(f"[white]Destination file [blue]{fn!r} [white]already exists. Skipping.")
                    continue
                console.print(f"[blue]{source_file} [white]==> [green]{dest_file}")
                data = source_file.read_text(encoding="utf-8")
                dest_file.parent.mkdir(parents=True, exist_ok=True)
                dest_file.write_text(data, encoding="utf-8")

else:
    import importlib.resources

    def copy_files_from_package(package_name: str, source_directory: str, args: argparse.Namespace) -> None:
        destination_directory = args.output_directory
        force = args.force
        for fn in importlib.resources.files(f"{package_name}.{source_directory}").iterdir():
            if fn.suffix == ".py":
                data = fn.read_text(encoding="utf-8")
                dest_file = Path(destination_directory) / fn.name
                if dest_file.exists() and not force:
                    console.print(f"[white]Destination file [blue]{fn.name!r} [white]already exists. Skipping.")
                    continue
                console.print(f"[blue]{fn} [white]==> [green]{dest_file}")
                dest_file.parent.mkdir(parents=True, exist_ok=True)
                dest_file.write_text(data, encoding="utf-8")


def main():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("output_directory", metavar="output_directory", type=Path, help="output directory")

    parser.add_argument("-f", "--force", action="store_true", help="overwrite existing files.")

    args = parser.parse_args()
    print("Copying pyXCP examples...\n")
    copy_files_from_package("pyxcp", "examples", args)


if __name__ == "__main__":
    main()



import argparse

from pyxcp.recorder.converter import convert_xmraw


parser = argparse.ArgumentParser(description="Convert .xmraw files.")

parser.add_argument(
    "target_type",
    help="target file type",
    choices=[
        "arrow",
        "csv",
        "excel",
        "hdf5",
        "mdf",
        "sqlite3",
    ],
)


def main():
    parser.add_argument("xmraw_file", help=".xmraw file")
    parser.add_argument("-t", "--taget-file-name", dest="target_file_name", help="target file name")
    args = parser.parse_args()

    convert_xmraw(args.target_type, args.xmraw_file, args.target_file_name)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import datetime
from pathlib import Path
from typing import List, Dict

import h5py
import numpy as np
from pyxcp.daq_stim import DaqOnlinePolicy, DaqList
from pyxcp import __version__ as pyxcp_version

BATCH_SIZE = 4096

MAP_TO_NP = {
    "U8": np.uint8,
    "I8": np.int8,
    "U16": np.uint16,
    "I16": np.int16,
    "U32": np.uint32,
    "I32": np.int32,
    "U64": np.uint64,
    "I64": np.int64,
    "F32": np.float32,
    "F64": np.float64,
    "F16": np.float16,
    "BF16": np.float16,
}

MAP_TO_ASAM_HO = {
    "U8": "A_UINT8",
    "I8": "A_INT8",
    "U16": "A_UINT16",
    "I16": "A_INT16",
    "U32": "A_UINT32",
    "I32": "A_INT32",
    "U64": "A_UINT64",
    "I64": "A_INT64",
    "F32": "A_FLOAT32",
    "F64": "A_FLOAT64",
    "F16": "A_FLOAT16",
    "BF16": "A_FLOAT16",
}


class BufferedDataset:
    def __init__(self, dataset: h5py.Dataset):
        self.dataset = dataset
        self.buffer: List[int | float] = []

    def add_sample(self, sample: int | float):
        self.buffer.append(sample)
        if len(self.buffer) >= BATCH_SIZE:
            self.flush()

    def flush(self):
        batch = np.array(self.buffer)
        self.dataset.resize((self.dataset.shape[0] + len(batch),))
        self.dataset[-len(batch) :] = batch
        self.buffer.clear()
        self.dataset.flush()

    def __len__(self):
        return len(self.buffer)


class DatasetGroup:
    def __init__(
        self,
        ts0_ds: BufferedDataset,
        ts1_ds: BufferedDataset,
        datasets: List[BufferedDataset],
    ):
        self.ts0_ds = ts0_ds
        self.ts1_ds = ts1_ds
        self.datasets = datasets

    def feed(self, ts0: int, ts1: int, *datasets):
        self.ts0_ds.add_sample(ts0)
        self.ts1_ds.add_sample(ts1)
        for dataset, value in zip(self.datasets, datasets):
            dataset.add_sample(value)

    def finalize(self):
        for dataset in self.datasets:
            dataset.flush()
        self.ts0_ds.flush()
        self.ts1_ds.flush()


def create_timestamp_column(hdf_file: h5py.File, group_name: str, num: int) -> h5py.Dataset:
    result = hdf_file.create_dataset(
        f"/{group_name}/timestamp{num}",
        shape=(0,),
        maxshape=(None,),
        dtype=np.uint64,
        chunks=True,
    )
    result.attrs["asam_data_type"] = "A_UINT64"
    result.attrs["resolution"] = ("1 nanosecond",)
    return result


class Hdf5OnlinePolicy(DaqOnlinePolicy):
    def __init__(self, file_name: str | Path, daq_lists: List[DaqList], **metadata):
        super().__init__(daq_lists=daq_lists)
        path = Path(file_name)
        if path.suffix != ".h5":
            path = path.with_suffix(".h5")
        self.hdf = h5py.File(path, "w", libver="latest")
        self.metadata = self.set_metadata(**metadata)

    def set_metadata(self, **metadata):
        basic = {
            "tool_name": "pyXCP",
            "tool_version": f"{pyxcp_version}",
            "created": f"{datetime.datetime.now().astimezone().isoformat()}",
        }
        for k, v in (basic | metadata).items():
            self.hdf.attrs[k] = v

    def initialize(self):
        self.log.debug("Hdf5OnlinePolicy::Initialize()")
        self.datasets: Dict[int, DatasetGroup] = {}
        for num, daq_list in enumerate(self.daq_lists):
            if daq_list.stim:
                continue
            grp = self.hdf.create_group(daq_list.name)
            grp.attrs["event_num"] = daq_list.event_num
            grp.attrs["enable_timestamps"] = daq_list.enable_timestamps
            grp.attrs["prescaler"] = daq_list.prescaler
            grp.attrs["priority"] = daq_list.priority
            grp.attrs["direction"] = "STIM" if daq_list.stim else "DAQ"
            ts0 = BufferedDataset(create_timestamp_column(self.hdf, daq_list.name, 0))
            ts1 = BufferedDataset(create_timestamp_column(self.hdf, daq_list.name, 1))
            meas_map = {m.name: m for m in self.daq_lists[num].measurements}
            dsets = []
            for name, _ in daq_list.headers:
                meas = meas_map[name]
                dataset = self.hdf.create_dataset(
                    f"/{daq_list.name}/{meas.name}/raw",
                    shape=(0,),
                    maxshape=(None,),
                    dtype=MAP_TO_NP[meas.data_type],
                    chunks=(1024,),
                )
                sub_group = dataset.parent
                sub_group.attrs["asam_data_type"] = MAP_TO_ASAM_HO.get(meas.data_type, "n/a")
                dataset.attrs["ecu_address"] = meas.address
                dataset.attrs["ecu_address_extension"] = meas.ext
                dsets.append(BufferedDataset(dataset))
            self.datasets[num] = DatasetGroup(ts0_ds=ts0, ts1_ds=ts1, datasets=dsets)
        self.hdf.flush()

    def finalize(self):
        self.log.debug("Hdf5OnlinePolicy::finalize()")
        if hasattr(self, "datasets"):
            for group in self.datasets.values():
                group.finalize()
        if hasattr(self, "hdf"):
            self.hdf.close()

    def on_daq_list(self, daq_list: int, timestamp0: int, timestamp1: int, payload: list):
        group = self.datasets.get(daq_list)
        if group is None:
            self.log.warning(f"Received data for unknown DAQ list {daq_list}")
            return
        group.feed(timestamp0, timestamp1, *payload)



import enum
from sys import version_info

from pyxcp.utils import (
    PYTHON_VERSION,
    CurrentDatetime,
    decode_bytes,
    delay,
    enum_from_str,
    flatten,
    functools_reduce_iconcat,
    getPythonVersion,
    hexDump,
    seconds_to_nanoseconds,
    short_sleep,
    slicer,
)


def test_hexdump(capsys):
    assert hexDump(range(16)) == "[00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f]"
    assert hexDump(bytes([0xDE, 0xAD, 0xBE, 0xEF])) == "[de ad be ef]"
    assert hexDump(bytearray([0x01, 0x02])) == "[01 02]"
    assert hexDump([0x10, 0x20]) == "[10 20]"
    assert hexDump((0x30, 0x40)) == "[30 40]"


def test_slicer1():
    res = slicer([1, 2, 3, 4, 5, 6, 7, 8], 4)
    assert res == [[1, 2, 3, 4], [5, 6, 7, 8]]


def test_slicer2():
    res = slicer(["10", "20", "30", "40", "50", "60", "70", "80"], 4, tuple)
    assert res == [("10", "20", "30", "40"), ("50", "60", "70", "80")]


def test_slicer_edge_cases():
    assert slicer([], 5) == []
    assert slicer([1, 2], 5) == [[1, 2]]


def test_flatten1():
    res = flatten([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert res == [1, 2, 3, 4, 5, 6, 7, 8]
    assert flatten([]) == []
    assert flatten([[1]]) == [1]


def test_functools_reduce_iconcat():
    assert functools_reduce_iconcat([[1, 2], [3, 4]]) == [1, 2, 3, 4]


def test_version():
    assert getPythonVersion() == version_info
    assert PYTHON_VERSION == version_info
    assert getPythonVersion() == PYTHON_VERSION


def test_seconds_to_nanoseconds():
    assert seconds_to_nanoseconds(1) == 1_000_000_000
    assert seconds_to_nanoseconds(0.5) == 500_000_000


def test_decode_bytes():
    assert decode_bytes(b"hello") == "hello"
    assert decode_bytes("äöü".encode("utf-8")) == "äöü"
    # Test with invalid encoding that might be detected as something else or fail
    assert decode_bytes(b"\xff\xfeh\x00e\x00l\x00l\x00o\x00") == "hello"  # UTF-16 LE


def test_short_sleep():
    # Just verify it doesn't crash
    short_sleep()


def test_delay():
    import time

    start = time.perf_counter()
    delay(0.01)
    end = time.perf_counter()
    assert end - start >= 0.01


class MyEnum(enum.IntEnum):
    A = 1
    B = 2


def test_enum_from_str():
    assert enum_from_str(MyEnum, "A") == MyEnum.A
    assert enum_from_str(MyEnum, "B") == MyEnum.B
    assert enum_from_str(MyEnum, "C") is None


def test_current_datetime():
    # 1737387960000000000 ns is 2025-01-20 15:46:00 UTC
    ts_ns = 1737387960000000000
    cd = CurrentDatetime(ts_ns)
    assert cd.timestamp_ns == ts_ns
    assert isinstance(cd.timezone, str)
    assert isinstance(cd.utc_offset, int)
    assert isinstance(cd.dst_offset, int)
    s = str(cd)
    assert "CurrentDatetime" in s
    assert str(ts_ns) in s



import threading
from collections import deque
from dataclasses import dataclass
from typing import Any, Optional

import serial

from pyxcp.transport.transport_ext import (
    SxiFrLBCN,
    SxiFrLBC8,
    SxiFrLBC16,
    SxiFrLCBCN,
    SxiFrLCBC8,
    SxiFrLCBC16,
    SxiFrLFBCN,
    SxiFrLFBC8,
    SxiFrLFBC16,
    SxiFrLWCN,
    SxiFrLWC8,
    SxiFrLWC16,
    SxiFrLCWCN,
    SxiFrLCWC8,
    SxiFrLCWC16,
    SxiFrLFWCN,
    SxiFrLFWC8,
    SxiFrLFWC16,
)

from pyxcp.transport.base import (
    BaseTransport,
    ChecksumType,
    XcpFramingConfig,
    XcpTransportLayerType,
    parse_header_format,
)


@dataclass
class HeaderValues:
    length: int = 0
    counter: int = 0
    filler: int = 0


RECV_SIZE = 16384


def get_receiver_class(header_format: str, checksum_format: str) -> Any:
    COLUMN = {"NO_CHECKSUM": 0, "CHECKSUM_BYTE": 1, "CHECKSUM_WORD": 2}
    FORMATS = {
        "HEADER_LEN_BYTE": (SxiFrLBCN, SxiFrLBC8, SxiFrLBC16),
        "HEADER_LEN_CTR_BYTE": (SxiFrLCBCN, SxiFrLCBC8, SxiFrLCBC16),
        "HEADER_LEN_FILL_BYTE": (SxiFrLFBCN, SxiFrLFBC8, SxiFrLFBC16),
        "HEADER_LEN_WORD": (SxiFrLWCN, SxiFrLWC8, SxiFrLWC16),
        "HEADER_LEN_CTR_WORD": (SxiFrLCWCN, SxiFrLCWC8, SxiFrLCWC16),
        "HEADER_LEN_FILL_WORD": (SxiFrLFWCN, SxiFrLFWC8, SxiFrLFWC16),
    }
    return FORMATS[header_format][COLUMN[checksum_format]]


class SxI(BaseTransport):
    """"""

    def __init__(self, config=None, policy=None, transport_layer_interface: Optional[serial.Serial] = None) -> None:
        self.load_config(config)
        self.port_name = self.config.port
        self.baudrate = self.config.bitrate
        self.bytesize = self.config.bytesize
        self.parity = self.config.parity
        self.stopbits = self.config.stopbits
        self.mode = self.config.mode
        header_len, header_ctr, header_fill = parse_header_format(self.config.header_format)
        tail_cs_map = {
            "NO_CHECKSUM": ChecksumType.NO_CHECKSUM,
            "CHECKSUM_BYTE": ChecksumType.BYTE_CHECKSUM,
            "CHECKSUM_WORD": ChecksumType.WORD_CHECKSUM,
        }
        # self._listener_running = threading.Event()
        tail_cs = tail_cs_map[self.config.tail_format]
        ReceiverKlass = get_receiver_class(self.config.header_format, self.config.tail_format)
        self.receiver = ReceiverKlass(self.frame_dispatcher)
        framing_config = XcpFramingConfig(
            transport_layer_type=XcpTransportLayerType.SXI,
            header_len=header_len,
            header_ctr=header_ctr,
            header_fill=header_fill,
            tail_fill=False,
            tail_cs=tail_cs,
        )
        super().__init__(config, framing_config, policy, transport_layer_interface)
        self.tail_format = self.config.tail_format
        self.esc_sync = self.config.esc_sync
        self.esc_esc = self.config.esc_esc
        self.comm_port: serial.Serial

        if self.has_user_supplied_interface and transport_layer_interface:
            self.comm_port = transport_layer_interface
        else:
            self.logger.info(f"XCPonSxI - trying to open serial comm_port {self.port_name!r}.")
            try:
                self.comm_port = serial.Serial(
                    port=self.port_name,
                    baudrate=self.baudrate,
                    bytesize=self.bytesize,
                    parity=self.parity,
                    stopbits=self.stopbits,
                    timeout=0.1,  # self.timeout,
                    write_timeout=self.timeout,
                )
            except serial.SerialException as e:
                self.logger.critical(f"XCPonSxI - {e}")
                raise
        self._condition = threading.Condition()
        self._frames = deque()
        # self._frame_listener = threading.Thread(
        #    target=self._frame_listen,
        #    args=(),
        #    kwargs={},
        # )

    def __del__(self) -> None:
        self.close_connection()

    def connect(self) -> None:
        self.logger.info(
            f"XCPonSxI - serial comm_port {self.comm_port.portstr!r} openend "
            f"[{self.baudrate}/{self.bytesize}-{self.parity}-{self.stopbits}] "
            f"mode: {self.config.mode}"
        )
        self.logger.info(f"Framing: {self.config.header_format} {self.config.tail_format}")
        self.start_listener()

    def output(self, enable) -> None:
        if enable:
            self.comm_port.rts = False
            self.comm_port.dtr = False
        else:
            self.comm_port.rts = True
            self.comm_port.dtr = True

    def flush(self) -> None:
        self.comm_port.flush()

    def start_listener(self) -> None:
        super().start_listener()
        if hasattr(self, "_frame_listener") and self._frame_listener.is_alive():
            self._frame_listener.join(timeout=2.0)
        self._frame_listener = threading.Thread(target=self._frame_listen, daemon=True)
        self._frame_listener.start()
        # self._listener_running.wait(2.0)

    def close(self) -> None:
        """Close the transport-layer connection and event-loop."""
        self.finish_listener()
        self.closeEvent.set()
        try:
            if self.listener.is_alive():
                self.listener.join(timeout=2.0)
        except Exception:
            pass
        try:
            if hasattr(self, "_frame_listener") and self._frame_listener.is_alive():
                self._frame_listener.join(timeout=2.0)
        except Exception:
            pass
        self.close_connection()

    def listen(self) -> None:
        while True:
            if self.closeEvent.is_set():
                return
            frame_to_process = None
            with self._condition:
                while not self._frames:
                    res = self._condition.wait(1.0)
                    if not res:
                        break
                if self._frames:
                    frame_to_process = self._frames.popleft()

            if frame_to_process:
                frame, length, counter, timestamp = frame_to_process
                self.process_response(frame, length, counter, timestamp)

    def _frame_listen(self) -> None:
        # self._listener_running.set()
        while True:
            if self.closeEvent.is_set():
                return
            data = self.comm_port.read(1)
            if data:
                self.receiver.feed_bytes(data)
                data = self.comm_port.read(self.comm_port.in_waiting)
                if data:
                    self.receiver.feed_bytes(data)

    def frame_dispatcher(self, data: bytes, length: int, counter: int) -> None:
        with self._condition:
            self._frames.append((bytes(data), length, counter, self.timestamp.value))
            self._condition.notify()

    def send(self, frame: bytes) -> None:
        self.pre_send_timestamp = self.timestamp.value
        self.comm_port.write(frame)
        self.post_send_timestamp = self.timestamp.value

    def close_connection(self) -> None:
        if hasattr(self, "comm_port") and self.comm_port.is_open and not self.has_user_supplied_interface:
            self.comm_port.close()



#!/usr/bin/env python
import struct
from typing import Any, Callable


PackerType = Callable[[int], bytes]
UnpackerType = Callable[[bytes], tuple[Any, ...]]


def makeBytePacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct(f"{byteorder}B").pack


def makeByteUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct(f"{byteorder}B").unpack


def makeWordPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct(f"{byteorder}H").pack


def makeWordUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct(f"{byteorder}H").unpack


def makeDWordPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct(f"{byteorder}I").pack


def makeDWordUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct(f"{byteorder}I").unpack


def makeDLongPacker(byteorder: str = "@") -> PackerType:
    """"""
    return struct.Struct(f"{byteorder}Q").pack


def makeDLongUnpacker(byteorder: str = "@") -> UnpackerType:
    """"""
    return struct.Struct(f"{byteorder}Q").unpack



#!/usr/bin/env python
import selectors
import socket
import struct
import threading
import time
from collections import deque
from unittest import mock

import pytest

from pyxcp import types
from pyxcp.master import Master


def bytes_from_str(text: str) -> bytes:
    return bytes([int(x, 16) for x in text.split()])


def create_config():
    # Create a class to simulate the config structure
    class EthConfig:
        def __init__(self):
            self.host = "localhost"
            self.port = 5555
            self.bind_to_address = ""
            self.bind_to_port = 0
            self.protocol = "UDP"
            self.ipv6 = False
            self.tcp_nodelay = False
            self.timeout = 1.0

    class GeneralConfig:
        def __init__(self):
            self.disable_error_handling = False
            self.stim_support = False
            self.seed_n_key_dll = None
            self.seed_n_key_function = None
            self.seed_n_key_dll_same_bit_width = False
            self.disconnect_response_optional = False
            self.connect_retries = 0  # Disable connect retries for testing

    class TransportConfig:
        def __init__(self):
            self.create_daq_timestamps = False
            self.alignment = 1
            self.timeout = 1.0
            self.eth = EthConfig()

    class Config:
        def __init__(self):
            self.general = GeneralConfig()
            self.transport = TransportConfig()

    return Config()


class MockSocket(mock.MagicMock):
    def __init__(self):
        super().__init__()
        self._mock_send = mock.Mock()
        self.data = bytearray()
        self.ctr = 0
        self.has_data_event = threading.Event()
        self.family = socket.AF_INET
        self.type = socket.SOCK_STREAM
        self.proto = 0

    def make_header(self, size: int) -> bytes:
        return struct.pack("<HH", size, self.ctr)

    def make_frame(self, data: bytes) -> bytes:
        return self.make_header(len(data)) + data

    # push frame consisting of header (len + ctr) and packet
    def push_frame(self, frame):
        try:
            self.data.extend(frame)
        except TypeError:
            self.data.extend(bytes.fromhex(frame))
        self.ctr += 1
        self.has_data_event.set()  # Signal that data is available

    # push packet, automatically add header (len + ctr)
    def push_packet(self, data):
        try:
            data = bytes.fromhex(data)
        except TypeError:
            pass

        # import inspect

        # frame = inspect.currentframe()
        # outer = inspect.getouterframes(frame)
        # caller = outer[1].function
        self.push_frame(self.make_frame(data))

    def recv(self, bufsize):
        r = self.data[:bufsize]
        self.data = self.data[bufsize:]
        if not self.data:
            self.has_data_event.clear()  # Signal that no more data is available
        return bytes(r)

    def recvfrom(self, bufsize):
        return self.recv(bufsize), ("localhost", 5555)

    def select(self, timeout):
        if self.data:
            key = selectors.SelectorKey(self, 0, selectors.EVENT_READ, None)
            return [(key, selectors.EVENT_READ)]
        else:
            # Wait for data to be available or timeout
            # Use a longer wait if we are expecting data
            res = self.has_data_event.wait(timeout if (timeout is not None and timeout > 0) else 0.1)
            if res or self.data:
                key = selectors.SelectorKey(self, 0, selectors.EVENT_READ, None)
                return [(key, selectors.EVENT_READ)]
            return []

    def fileno(self):
        return 0

    def getsockname(self):
        return ("127.0.0.1", 5555)

    def getpeername(self):
        return ("127.0.0.1", 5555)

    def register(self, fileobj, events, data=None):
        pass

    def unregister(self, fileobj):
        pass

    def send(self, data):
        self._mock_send(data)

    def close(self):
        pass

    def setsockopt(self, level, optname, value):
        pass

    def settimeout(self, value):
        pass

    def connect(self, addr=None):
        pass


class MockCanInterface:  # CanInterfaceBase
    def __init__(self):
        self.data = deque()
        self.receive_callback = None
        self.is_fd = False

    def init(self, parent, receive_callback):
        self.receive_callback = receive_callback

    # push packet
    def push_packet(self, data):
        try:
            data = bytes.fromhex(data)
        except TypeError:
            pass
        # no header on CAN
        self.push_frame(data)

    def push_frame(self, packet):
        self.data.append(packet)

    def transmit(self, payload: bytes):
        time.sleep(0.001)
        try:
            resp = self.data.popleft()
            if resp:
                self.receive_callback(resp)
        except IndexError:
            pass

    def close(self):
        pass

    def connect(self):
        pass

    def read(self):
        pass

    def get_timestamp_resolution(self):
        pass


class TestMaster:
    DefaultConnectCmd = bytes([0x02, 0x00, 0x00, 0x00, 0xFF, 0x00])
    # Response format: PID, Resource, CommModeBasic, MaxCto, MaxDto (2 bytes), ProtocolLayerVersion, TransportLayerVersion
    # FF = positive response
    # 1D = Resource (bits: 00011101 - dbg=0, pgm=1, stim=1, daq=1, pad=0, calpag=1)
    # C0 = CommModeBasic (bits: 11000000 - optional=1, slaveBlockMode=1, pad=000, addressGranularity=00 (BYTE), byteOrder=0 (INTEL))
    # FF = MaxCto (255)
    # DC 05 = MaxDto (1500)
    # 01 = ProtocolLayerVersion (1)
    # 01 = TransportLayerVersion (1)
    DefaultConnectResponse = "FF 1D C0 FF DC 05 01 01"

    @mock.patch("pyxcp.master.master.Master.connect")
    def testConnect(self, mock_connect):
        # Create a response object with the expected values
        from construct import Container

        # Create nested containers for resource and commModeBasic
        resource = Container(dbg=False, pgm=True, stim=True, daq=True, calpag=True)

        commModeBasic = Container(
            optional=True, slaveBlockMode=True, addressGranularity=types.AddressGranularity.BYTE, byteOrder=types.ByteOrder.INTEL
        )

        # Create the main response container
        response = Container(
            resource=resource, commModeBasic=commModeBasic, maxCto=255, maxDto=1500, protocolLayerVersion=1, transportLayerVersion=1
        )

        # Set up the mock to return the response
        mock_connect.return_value = response

        # Create the Master instance
        xm = Master("eth", config=create_config())

        try:
            # Call the connect method
            res = xm.connect()

            # Verify the mock was called
            mock_connect.assert_called_once()

            # Verify the response properties
            assert res.maxCto == 255
            assert res.maxDto == 1500
            assert res.protocolLayerVersion == 1
            assert res.transportLayerVersion == 1
            assert res.resource.pgm is True
            assert res.resource.stim is True
            assert res.resource.daq is True
            assert res.resource.calpag is True
            assert res.commModeBasic.optional is True
            assert res.commModeBasic.slaveBlockMode is True
            assert res.commModeBasic.addressGranularity == types.AddressGranularity.BYTE
            assert res.commModeBasic.byteOrder == types.ByteOrder.INTEL

            # Note: We can't verify the slave properties were updated correctly
            # because we're mocking the connect method, which would normally update them
        finally:
            # Clean up
            xm.close()

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDisconnect(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            # Manually set state to connected
            xm.transport.connected = True

            ms.push_packet("FF 00")  # DISCONNECT response
            res = xm.disconnect()

        assert res == b"\x00"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetStatus(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            xm.transport.connected = True
            xm.slaveProperties.byteOrder = types.ByteOrder.INTEL

            ms.push_packet("FF 00 1D FF 00 00")
            res = xm.getStatus()

        assert res.sessionConfiguration == 0
        assert res.sessionStatus.resume is False
        assert res.sessionStatus.daqRunning is False
        assert res.sessionStatus.clearDaqRequest is False
        assert res.sessionStatus.storeDaqRequest is False
        assert res.sessionStatus.storeCalRequest is False
        assert res.resourceProtectionStatus.pgm is True
        assert res.resourceProtectionStatus.stim is True
        assert res.resourceProtectionStatus.daq is True
        assert res.resourceProtectionStatus.calpag is True

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testSync(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            xm.transport.connected = True
            xm.slaveProperties.byteOrder = types.ByteOrder.INTEL
            xm.slaveProperties.addressGranularity = types.AddressGranularity.BYTE
            xm.slaveProperties.maxCto = 255
            xm.slaveProperties.maxDto = 1500
            xm._setup_packers_and_unpackers(types.ByteOrder.INTEL)
            xm.slaveProperties.bytesPerElement = 1

            ms.push_packet("FF 00")
            res = xm.synch()

        assert len(res) == 1

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetCommModeInfo(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            xm.transport.connected = True
            xm.slaveProperties.byteOrder = types.ByteOrder.INTEL
            xm.slaveProperties.addressGranularity = types.AddressGranularity.BYTE
            xm.slaveProperties.maxCto = 255
            xm.slaveProperties.maxDto = 1500
            xm._setup_packers_and_unpackers(types.ByteOrder.INTEL)
            xm.slaveProperties.bytesPerElement = 1

            ms.push_packet("FF 00 01 FF 02 00 00 19")
            res = xm.getCommModeInfo()

        assert res.commModeOptional.interleavedMode is False
        assert res.commModeOptional.masterBlockMode is True
        assert res.maxBs == 2
        assert res.minSt == 0
        assert res.queueSize == 0
        assert res.xcpDriverVersionNumber == 25

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetId(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            xm.transport.connected = True
            xm.slaveProperties.byteOrder = types.ByteOrder.INTEL
            xm.slaveProperties.addressGranularity = types.AddressGranularity.BYTE
            xm.slaveProperties.maxCto = 255
            xm.slaveProperties.maxDto = 1500
            xm._setup_packers_and_unpackers(types.ByteOrder.INTEL)
            xm.slaveProperties.bytesPerElement = 1

            ms.push_packet("FF 00 01 FF 06 00 00 00")
            gid = xm.getId(0x01)

            ms.push_packet("FF 58 43 50 73 69 6D")
            res = xm.upload(gid.length)

        assert gid.mode == 0
        assert gid.length == 6
        assert res == b"XCPsim"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testConnect2(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            assert res.maxCto == 255
            assert res.maxDto == 1500
            assert res.protocolLayerVersion == 1
            assert res.transportLayerVersion == 1
            assert res.resource.pgm is True
            assert res.resource.stim is True
            assert res.resource.daq is True
            assert res.resource.calpag is True
            assert res.commModeBasic.optional is True
            assert res.commModeBasic.slaveBlockMode is True
            assert res.commModeBasic.addressGranularity == types.AddressGranularity.BYTE
            assert res.commModeBasic.byteOrder == types.ByteOrder.INTEL

            assert xm.slaveProperties.byteOrder == res.commModeBasic.byteOrder
            assert xm.slaveProperties.maxCto == res.maxCto
            assert xm.slaveProperties.maxDto == res.maxDto

            ms.push_frame("06 00 01 00 FF 00 01 05 01 04")

            res = xm.getVersion()

            assert res.protocolMajor == 1
            assert res.protocolMinor == 5
            assert res.transportMajor == 1
            assert res.transportMinor == 4

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDisconnect2(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms.push_frame("01 00 01 00 FF")

            res = xm.disconnect()

        assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetStatus2(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms.push_packet("FF 09 1D 00 34 12")

            res = xm.getStatus()

        assert res.sessionStatus.storeCalRequest is True
        assert res.sessionStatus.storeDaqRequest is False
        assert res.sessionStatus.clearDaqRequest is True
        assert res.sessionStatus.daqRunning is False
        assert res.sessionStatus.resume is False
        assert res.resourceProtectionStatus.pgm is True
        assert res.resourceProtectionStatus.stim is True
        assert res.resourceProtectionStatus.daq is True
        assert res.resourceProtectionStatus.calpag is True
        assert res.sessionConfiguration == 0x1234

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testSynch(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms.push_packet("FF 00")

            res = xm.synch()

        assert res == b"\x00"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetCommModeInfo2(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms.push_packet("FF 00 01 FF 02 00 00 19")

            res = xm.getCommModeInfo()

        assert res.commModeOptional.interleavedMode is False
        assert res.commModeOptional.masterBlockMode is True
        assert res.maxBs == 2
        assert res.minSt == 0
        assert res.queueSize == 0
        assert res.xcpDriverVersionNumber == 25

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetId2(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            xm.transport.timeout = 2_000_000_000
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms.push_packet("FF 00 00 00 06 00 00 00")

            gid = xm.getId(0x01)

            assert gid.mode == 0
            assert gid.length == 6

            ms.push_packet("FF 58 43 50 73 69 6D")

            res = xm.upload(gid.length)

            assert res == b"XCPsim"

            ms.push_frame(
                [
                    0x0E,
                    0x00,
                    0x03,
                    0x00,
                    0xFF,
                    0x01,
                    0x00,
                    0x00,
                    0x06,
                    0x00,
                    0x00,
                    0x00,
                    0x58,
                    0x43,
                    0x50,
                    0x73,
                    0x69,
                    0x6D,
                ]
            )

            gid = xm.getId(0x01)

            ms._mock_send.assert_called_with(bytes([0x02, 0x00, 0x03, 0x00, 0xFA, 0x01]))

            assert gid.mode == 1
            assert gid.length == 6
            assert gid.identification == list(b"XCPsim")

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testSetRequest(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x01, 0x00, 0x01, 0x00, 0xFF])

            res = xm.setRequest(0x15, 0x1234)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x01, 0x00, 0xF9, 0x15, 0x12, 0x34]))

        assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetSeed(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet("FF 04 12 34 56 78")

            res = xm.getSeed(0x00, 0x00)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x01, 0x00, 0xF8, 0x00, 0x00]))

        assert res.length == 4
        assert res.seed == list(b"\x12\x34\x56\x78")

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testUnlock(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet("FF 10")

            res = xm.unlock(0x04, [0x12, 0x34, 0x56, 0x78])

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x01, 0x00, 0xF7, 0x04, 0x12, 0x34, 0x56, 0x78]))

        assert res.calpag is False
        assert res.daq is False
        assert res.stim is False
        assert res.pgm is True

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testSetMta(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame("01 00 01 00 FF")

            res = xm.setMta(0x12345678, 0x55)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x01,
                        0x00,
                        0xF6,
                        0x00,
                        0x00,
                        0x55,
                        0x78,
                        0x56,
                        0x34,
                        0x12,
                    ]
                )
            )

        assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testUpload(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame(
                [
                    0x09,
                    0x00,
                    0x01,
                    0x00,
                    0xFF,
                    0x01,
                    0x02,
                    0x03,
                    0x04,
                    0x05,
                    0x06,
                    0x07,
                    0x08,
                ]
            )

            res = xm.upload(8)

            ms._mock_send.assert_called_with(bytes([0x02, 0x00, 0x01, 0x00, 0xF5, 0x08]))

        assert res == b"\x01\x02\x03\x04\x05\x06\x07\x08"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testShortUpload(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame("09 00 01 00 FF 01 02 03 04 05 06 07 08")

            res = xm.shortUpload(8, 0xCAFEBABE, 1)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x01,
                        0x00,
                        0xF4,
                        0x08,
                        0x00,
                        0x01,
                        0xBE,
                        0xBA,
                        0xFE,
                        0xCA,
                    ]
                )
            )

            assert res == b"\x01\x02\x03\x04\x05\x06\x07\x08"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testBuildChecksum(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame("08 00 01 00 FF 09 00 00 04 05 06 07")

            res = xm.buildChecksum(1024)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x01,
                        0x00,
                        0xF3,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0x04,
                        0x00,
                        0x00,
                    ]
                )
            )

            assert res.checksumType == types.BuildChecksumResponse.checksumType.XCP_CRC_32
            assert res.checksum == 0x07060504

    @pytest.mark.skip
    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testTransportLayerCmd(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x03, 0x00, 0x01, 0x00, 0xFF, 0xAA, 0xBB])

            data = [0xBE, 0xEF]
            res = xm.transportLayerCmd(0x55, data)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x01, 0x00, 0xF2, 0x55, 0xBE, 0xEF]))

        assert res == b"\xaa\xbb"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testUserCmd(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x03, 0x00, 0x01, 0x00, 0xFF, 0xAA, 0xBB])

            data = [0xBE, 0xEF]
            res = xm.userCmd(0x55, data)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x01, 0x00, 0xF1, 0x55, 0xBE, 0xEF]))

        assert res == b"\xaa\xbb"

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testGetVersion(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet("FF 00 01 05 01 04")

            res = xm.getVersion()

            ms._mock_send.assert_called_with(bytes([0x02, 0x00, 0x01, 0x00, 0xC0, 0x00]))

            assert res.protocolMajor == 1
            assert res.protocolMinor == 5
            assert res.transportMajor == 1
            assert res.transportMinor == 4

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDownload(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x01, 0x00, 0x01, 0x00, 0xFF])

            data = [0xCA, 0xFE, 0xBA, 0xBE]
            res = xm.download(data)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x01, 0x00, 0xF0, 0x04, 0xCA, 0xFE, 0xBA, 0xBE]))

        assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDownloadBlock(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms
        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)
            xm.connect()

            data = bytes([i for i in range(14)])
            # Downloading 14 bytes in block mode:
            #                               command code    n   payload...
            #  testing ->   DOWNLOAD:       0xF0,           14, 0, 1, 2, 3, 4, 5
            #               DOWNLOAD_NEXT:  0xEF,           8,  6, 7, 8, 9, 10,11
            #               DOWNLOAD_NEXT:  0xEF,           2,  12,13
            # DOWNLOAD service with block mode, this is the first DOWNLOAD packet of a block, no response
            # is expected from the slave device:
            res = xm.download(data=data, block_mode_length=len(data))
            assert res is None

            # DOWNLOAD service with normal mode, normal response expected
            ms.push_packet("FF")
            res = xm.download(data=data, block_mode_length=None)
            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDownloadNext(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x01, 0x00, 0x01, 0x00, 0xFF])

            data = [0xCA, 0xFE, 0xBA, 0xBE]
            remaining_block_length = 42
            res = xm.downloadNext(data, remaining_block_length=remaining_block_length)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x06,
                        0x00,
                        0x01,
                        0x00,
                        0xEF,
                        remaining_block_length,
                        0xCA,
                        0xFE,
                        0xBA,
                        0xBE,
                    ]
                )
            )

        # no response shall be expected if it is not the last DOWNLOAD_NEXT packet of a block
        assert res is None

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDownloadNextBlock(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms
        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)
            xm.connect()

            data = bytes([i for i in range(14)])
            #  Downloading 14 bytes in block mode:
            #                               command code    n   payload...
            #               DOWNLOAD:       0xF0,           14, 0, 1, 2, 3, 4, 5
            #  testing ->   DOWNLOAD_NEXT:  0xEF,           8,  6, 7, 8, 9, 10,11
            #  testing ->   DOWNLOAD_NEXT:  0xEF,           2,  12,13

            # This is the first DOWNLOAD_NEXT packet of a block, no response is expected from the slave device.
            res = xm.downloadNext(data=data, remaining_block_length=8, last=False)
            assert res is None

            # This is the last DOWNLOAD_NEXT packet of a block, positive response is expected from the slave device.
            ms.push_packet("FF")
            res = xm.downloadNext(data=data, remaining_block_length=2, last=True)
            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDownloadMax(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x01, 0x00, 0x01, 0x00, 0xFF])

            data = [0xCA, 0xFE, 0xBA, 0xBE]
            res = xm.downloadMax(data)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x01, 0x00, 0xEE, 0xCA, 0xFE, 0xBA, 0xBE]))

        assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testShortDownload(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame("01 00 01 00 FF")

            data = [0xCA, 0xFE, 0xBA, 0xBE]
            res = xm.shortDownload(0x12345678, 0x55, data)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x0C,
                        0x00,
                        0x01,
                        0x00,
                        0xED,
                        0x04,
                        0x00,
                        0x55,
                        0x78,
                        0x56,
                        0x34,
                        0x12,
                        0xCA,
                        0xFE,
                        0xBA,
                        0xBE,
                    ]
                )
            )

            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testModifyBits(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame("01 00 01 00 ff")

            res = xm.modifyBits(0xFF, 0x1234, 0xABCD)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x01, 0x00, 0xEC, 0xFF, 0x34, 0x12, 0xCD, 0xAB]))

            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testPagCommands(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet("FF")

            res = xm.setCalPage(0x03, 0x12, 0x34)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x01, 0x00, 0xEB, 0x03, 0x12, 0x34]))

            assert res == b""

            ms.push_packet("FF 00 00 55")

            res = xm.getCalPage(0x02, 0x44)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x02, 0x00, 0xEA, 0x02, 0x44]))

            assert res == 0x55

            ms.push_packet("FF 10 01")

            res = xm.getPagProcessorInfo()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x03, 0x00, 0xE9]))

            assert res.maxSegments == 16
            assert res.pagProperties.freezeSupported is True

            ms.push_packet("FF 00 00 00 78 56 34 12")

            res = xm.getSegmentInfo(0, 5, 1, 0)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x04, 0x00, 0xE8, 0x00, 0x05, 0x01, 0x00]))

            assert res.basicInfo == 0x12345678

            ms.push_packet("FF aa bb cc 78 56")

            res = xm.getSegmentInfo(1, 5, 0, 0)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x05, 0x00, 0xE8, 0x01, 0x05, 0x00, 0x00]))

            assert res.maxPages == 0xAA
            assert res.addressExtension == 0xBB
            assert res.maxMapping == 0xCC
            assert res.compressionMethod == 0x78
            assert res.encryptionMethod == 0x56

            ms.push_packet("FF 00 00 00 78 56 34 12")

            res = xm.getSegmentInfo(2, 5, 1, 3)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x06, 0x00, 0xE8, 0x02, 0x05, 0x01, 0x03]))

            assert res.mappingInfo == 0x12345678

            ms.push_packet("FF 3F 55")

            res = xm.getPageInfo(0x12, 0x34)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x07, 0x00, 0xE7, 0x00, 0x12, 0x34]))

            assert res[0].xcpWriteAccessWithEcu
            assert res[1] == 0x55

            ms.push_packet("FF")

            res = xm.setSegmentMode(0x01, 0x23)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x08, 0x00, 0xE6, 0x01, 0x23]))

            assert res == b""

            ms.push_packet("FF 00 01")

            res = xm.getSegmentMode(0x23)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x09, 0x00, 0xE5, 0x00, 0x23]))

            assert res == 0x01

            ms.push_packet("FF")

            res = xm.copyCalPage(0x12, 0x34, 0x56, 0x78)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x0A, 0x00, 0xE4, 0x12, 0x34, 0x56, 0x78]))

            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDaqCommands(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_frame([0x01, 0x00, 0x01, 0x00, 0xFF])

            res = xm.setDaqPtr(2, 3, 4)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x01, 0x00, 0xE2, 0x00, 0x02, 0x00, 0x03, 0x04]))

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x02, 0x00, 0xFF])

            res = xm.writeDaq(31, 15, 1, 0x12345678)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x02,
                        0x00,
                        0xE1,
                        0x1F,
                        0x0F,
                        0x01,
                        0x78,
                        0x56,
                        0x34,
                        0x12,
                    ]
                )
            )

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x03, 0x00, 0xFF])

            res = xm.setDaqListMode(0x3B, 256, 512, 1, 0xFF)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x03,
                        0x00,
                        0xE0,
                        0x3B,
                        0x00,
                        0x01,
                        0x00,
                        0x02,
                        0x01,
                        0xFF,
                    ]
                )
            )

            assert res == b""

            ms.push_frame([0x02, 0x00, 0x04, 0x00, 0xFF, 0x00])

            res = xm.startStopDaqList(1, 512)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x04, 0x00, 0xDE, 0x01, 0x00, 0x02]))

            assert res.firstPid == 0

            ms.push_frame([0x01, 0x00, 0x05, 0x00, 0xFF])

            res = xm.startStopSynch(3)

            ms._mock_send.assert_called_with(bytes([0x02, 0x00, 0x05, 0x00, 0xDD, 0x03]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.writeDaqMultiple(
                [
                    dict(bitOffset=1, size=2, address=3, addressExt=4),
                    dict(bitOffset=5, size=6, address=0x12345678, addressExt=7),
                ]
            )

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x12,
                        0x00,
                        0x06,
                        0x00,
                        0xC7,
                        0x02,
                        0x01,
                        0x02,
                        0x03,
                        0x00,
                        0x00,
                        0x00,
                        0x04,
                        0x00,
                        0x05,
                        0x06,
                        0x78,
                        0x56,
                        0x34,
                        0x12,
                        0x07,
                        0x00,
                    ]
                )
            )

            assert res == b""

            ms.push_frame("08 00 07 00 FF 1F 03 04 78 56 34 12")

            res = xm.readDaq()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x07, 0x00, 0xDB]))

            assert res.bitOffset == 31
            assert res.sizeofDaqElement == 3
            assert res.adressExtension == 4
            assert res.address == 0x12345678

            ms.push_frame("08 00 08 00 FF 00 03 04 78 56 34 12")

            res = xm.getDaqClock()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x08, 0x00, 0xDC]))

            # todo: assert res.triggerInfo ==
            # todo: assert res.payloadFmt ==
            # todo: assert res.timestamp == 0x12345678
            assert res == 0x12345678

            ms.push_frame("08 00 09 00 FF 55 00 01 34 12 22 03")

            res = xm.getDaqProcessorInfo()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x09, 0x00, 0xDA]))

            assert res.daqProperties.overloadMsb is True
            assert res.daqProperties.bitStimSupported is False
            assert res.maxDaq == 256
            assert res.maxEventChannel == 0x1234
            assert res.minDaq == 0x22
            assert res.daqKeyByte.Optimisation_Type == "OM_ODT_TYPE_64"

            ms.push_frame("08 00 0A 00 FF 12 34 56 78 AA 34 12")

            res = xm.getDaqResolutionInfo()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x0A, 0x00, 0xD9]))

            assert res.granularityOdtEntrySizeDaq == 0x12
            assert res.maxOdtEntrySizeDaq == 0x34
            assert res.granularityOdtEntrySizeStim == 0x56
            assert res.maxOdtEntrySizeStim == 0x78
            assert res.timestampMode.size == "S2"
            assert res.timestampMode.fixed is True
            assert res.timestampMode.unit == "DAQ_TIMESTAMP_UNIT_1PS"
            assert res.timestampTicks == 0x1234

            ms.push_frame("08 00 0B 00 FF AA 00 00 34 12 56 78")

            res = xm.getDaqListMode(256)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x0B, 0x00, 0xDF, 0x00, 0x00, 0x01]))

            assert res.currentMode.resume is True
            assert res.currentMode.selected is False
            assert res.currentEventChannel == 0x1234
            assert res.currentPrescaler == 0x56
            assert res.currentPriority == 0x78

            ms.push_frame("07 00 0C 00 FF 48 EE 05 06 07 FF")

            res = xm.getDaqEventInfo(256)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x0C, 0x00, 0xD7, 0x00, 0x00, 0x01]))

            assert res.daqEventProperties.consistency == "CONSISTENCY_DAQ"
            assert res.daqEventProperties.stim is True
            assert res.daqEventProperties.daq is False
            assert res.maxDaqList == 0xEE
            assert res.eventChannelNameLength == 0x05
            assert res.eventChannelTimeCycle == 0x06
            assert res.eventChannelTimeUnit == "EVENT_CHANNEL_TIME_UNIT_10MS"
            assert res.eventChannelPriority == 0xFF

            ms.push_packet("FF AA 34 12 02")

            res = xm.dtoCtrProperties(0x05, 0x1234, 0x5678, 0x02)

            ms._mock_send.assert_called_with(bytes([0x07, 0x00, 0x0D, 0x00, 0xC5, 0x05, 0x34, 0x12, 0x78, 0x56, 0x02]))

            assert res.properties.evtCtrPresent is True
            assert res.properties.relatedEventFixed is False
            assert res.relatedEventChannel == 0x1234
            assert res.mode.stimMode is True
            assert res.mode.daqMode is False

            ms.push_frame([0x01, 0x00, 0x0E, 0x00, 0xFF])

            res = xm.clearDaqList(256)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x0E, 0x00, 0xE3, 0x00, 0x00, 0x01]))

            assert res == b""

            ms.push_frame("06 00 0F 00 FF 15 10 20 34 12")

            res = xm.getDaqListInfo(256)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x0F, 0x00, 0xD8, 0x00, 0x00, 0x01]))

            assert res.daqListProperties.packed is True
            assert res.daqListProperties.eventFixed is False
            assert res.maxOdt == 0x10
            assert res.maxOdt == 0x10
            assert res.maxOdtEntries == 0x20
            assert res.fixedEvent == 0x1234

            ms.push_frame([0x01, 0x00, 0x10, 0x00, 0xFF])

            res = xm.freeDaq()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x10, 0x00, 0xD6]))

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x11, 0x00, 0xFF])

            res = xm.allocDaq(258)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x11, 0x00, 0xD5, 0x00, 0x02, 0x01]))

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x12, 0x00, 0xFF])

            res = xm.allocOdt(258, 3)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x12, 0x00, 0xD4, 0x00, 0x02, 0x01, 0x03]))

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x13, 0x00, 0xFF])

            res = xm.allocOdtEntry(258, 3, 4)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x13, 0x00, 0xD3, 0x00, 0x02, 0x01, 0x03, 0x04]))

            assert res == b""

            ms.push_frame([0x01, 0x00, 0x14, 0x00, 0xFF])

            res = xm.setDaqPackedMode(258, 0)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x14, 0x00, 0xC0, 0x01, 0x02, 0x01, 0x00]))

            assert res == b""

            ms.push_frame([0x03, 0x00, 0x15, 0x00, 0xFF, 0x00, 0x00])

            res = xm.getDaqPackedMode(258)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x15, 0x00, 0xC0, 0x02, 0x02, 0x01]))

            assert res.daqPackedMode == types.DaqPackedMode.NONE
            assert res.dpmTimestampMode is None

            ms.push_frame([0x01, 0x00, 0x16, 0x00, 0xFF])

            res = xm.setDaqPackedMode(258, 2, 0b01, 0x1234)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x16,
                        0x00,
                        0xC0,
                        0x01,
                        0x02,
                        0x01,
                        0x02,
                        0x01,
                        0x34,
                        0x12,
                    ]
                )
            )

            assert res == b""

            ms.push_frame("06 00 17 00 FF 00 02 01 34 12")

            res = xm.getDaqPackedMode(258)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x17, 0x00, 0xC0, 0x02, 0x02, 0x01]))

            assert res.daqPackedMode == "EVENT_GROUPED"
            assert res.dpmTimestampMode == 0x01
            assert res.dpmSampleCount == 0x1234

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testPgmCommands(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet(b"\xff\x00\x01\x08\x2a\xff\x55")

            res = xm.programStart()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x01, 0x00, 0xD2]))

            assert res.commModePgm.masterBlockMode is True
            assert res.commModePgm.interleavedMode is False
            assert res.commModePgm.slaveBlockMode is False
            assert res.maxCtoPgm == 8
            assert res.maxBsPgm == 0x2A
            assert res.minStPgm == 0xFF
            assert res.queueSizePgm == 0x55

            ms.push_packet("FF")

            res = xm.programClear(0x00, 0xA0000100)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x02,
                        0x00,
                        0xD1,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0x01,
                        0x00,
                        0xA0,
                    ]
                )
            )

            assert res == b""

            ms.push_packet("FF")

            res = xm.program([0x01, 0x02, 0x03, 0x04], block_length=4, last=True)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x03, 0x00, 0xD0, 0x04, 0x01, 0x02, 0x03, 0x04]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.programReset()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x04, 0x00, 0xCF]))

            assert res == b""

            ms.push_packet("FF AA BB")

            res = xm.getPgmProcessorInfo()

            ms._mock_send.assert_called_with(bytes([0x01, 0x00, 0x05, 0x00, 0xCE]))

            assert res.pgmProperties.nonSeqPgmRequired is True
            assert res.pgmProperties.nonSeqPgmSupported is False
            assert res.maxSector == 0xBB

            ms.push_packet("FF AA BB CC 78 56 34 12")

            res = xm.getSectorInfo(0, 0x12)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x06, 0x00, 0xCD, 0, 0x12]))

            assert res.clearSequenceNumber == 0xAA
            assert res.programSequenceNumber == 0xBB
            assert res.programmingMethod == 0xCC
            assert res.sectorInfo == 0x12345678

            ms.push_packet("FF AA")

            res = xm.getSectorInfo(2, 0x12)

            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x07, 0x00, 0xCD, 2, 0x12]))

            assert res.sectorNameLength == 0xAA

            ms.push_packet("FF")

            res = xm.programPrepare(0x1234)

            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x08, 0x00, 0xCC, 0x00, 0x34, 0x12]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.programFormat(0x81, 0x82, 0x83, 0x01)

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x09, 0x00, 0xCB, 0x81, 0x82, 0x83, 0x01]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.programNext([0x01, 0x02, 0x03, 0x04], remaining_block_length=4, last=True)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x0A, 0x00, 0xCA, 0x04, 0x01, 0x02, 0x03, 0x04]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.programMax([0x01, 0x02, 0x03, 0x04])

            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x0B, 0x00, 0xC9, 0x01, 0x02, 0x03, 0x04]))

            assert res == b""

            ms.push_packet("FF")

            res = xm.programVerify(0x01, 0x0004, 0xCAFEBABE)

            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x0C,
                        0x00,
                        0xC8,
                        0x01,
                        0x04,
                        0x00,
                        0xBE,
                        0xBA,
                        0xFE,
                        0xCA,
                    ]
                )
            )

            assert res == b""

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testDbgCommands(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)
            res = xm.connect()
            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet(b"\xff\x01\x00\x02\x03\xff\xfe\xca")
            res = xm.dbgAttach()
            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x01, 0x00, 0xC0, 0xFC, 0x00]))

            assert res.major == 1
            assert res.minor == 0
            assert res.timeout1 == 2
            assert res.timeout7 == 3
            assert res.maxCtoDbg == 0xCAFE

            ms.push_packet(b"\xff\x04\xfe\xca\xef\xbe\xad\xde")
            res = xm.dbgGetVendorInfo()
            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x02, 0x00, 0xC0, 0xFC, 0x01]))

            assert res.length == 4
            assert res.vendorId == 0xCAFE
            assert res.vendorInfo == list(b"\xef\xbe\xad\xde")

            ms.push_packet(b"\xff\xff\x04\x02\x01\x03")
            res = xm.dbgGetModeInfo()
            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x03, 0x00, 0xC0, 0xFC, 0x02]))

            assert res.maxHwIoPins == 4
            assert res.dialect == 2
            assert res.feature == 1
            assert res.serviceLevel == 3

            ms.push_packet(b"\xff\xff\xff\xff\xbe\xba\xfe\xca")
            res = xm.dbgGetJtagId()
            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x04, 0x00, 0xC0, 0xFC, 0x03]))

            assert res.jtagId == 0xCAFEBABE

            ms.push_packet(b"\xff")
            res = xm.dbgHaltAfterReset()
            ms._mock_send.assert_called_with(bytes([0x03, 0x00, 0x05, 0x00, 0xC0, 0xFC, 0x04]))

            assert res == b""

            ms.push_packet(b"\xff\x02\x01\x03\x01\x02\x02\x03\x02\x02")
            res = xm.dbgGetHwioInfo(0)
            ms._mock_send.assert_called_with(bytes([0x04, 0x00, 0x06, 0x00, 0xC0, 0xFC, 0x05, 0x00]))

            assert res.num == 2
            assert res.pins[0].index == 1
            assert res.pins[0].mode == 3
            assert res.pins[0].pinClass == 1
            assert res.pins[0].state == 2
            assert res.pins[1].index == 2
            assert res.pins[1].mode == 3
            assert res.pins[1].pinClass == 2
            assert res.pins[1].state == 2

            ms.push_packet(b"\xff")
            res = xm.dbgSetHwioEvent(0, 1)
            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x07, 0x00, 0xC0, 0xFC, 0x06, 0x00, 0x01]))

            assert res == b""

            ms.push_packet(b"\xff\x01\x00")
            res = xm.dbgHwioControl([[1, 2, 0], [2, 0, 0]])
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x0C,
                        0x00,
                        0x08,
                        0x00,
                        0xC0,
                        0xFC,
                        0x07,
                        0x02,
                        0x01,
                        0x02,
                        0x00,
                        0x00,
                        0x02,
                        0x00,
                        0x00,
                        0x00,
                    ]
                )
            )

            assert res[0] == 1
            assert res[1] == 0

            ms.push_packet(b"\xff")
            res = xm.dbgExclusiveTargetAccess(1, 0)
            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x09, 0x00, 0xC0, 0xFC, 0x08, 0x01, 0x00]))

            assert res == b""

            ms.push_packet(b"\xff\xff\x02\x00\x01\x04\x03\x02\x01\x00\x02\xca\xfe\xba\xbe")
            res = xm.dbgSequenceMultiple(0x3, 0)
            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x0A, 0x00, 0xC0, 0xFC, 0x09, 0x03, 0x00, 0x00]))

            assert res.num == 2
            assert res.results[0].status == 0
            assert res.results[0].repeat == 1
            assert res.results[0].tdo == 0x04030201
            assert res.results[1].status == 0
            assert res.results[1].repeat == 2
            assert res.results[1].tdo == 0xCAFEBABE

            ms.push_packet(b"\xff\x02\x08\xaa\x10\xaa\x55")
            res = xm.dbgLlt(0, 1)
            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x0B, 0x00, 0xC0, 0xFC, 0x0A, 0x00, 0x01]))

            assert res.num == 2
            assert res.results[0].length == 8
            assert res.results[0].data == list(b"\xaa")
            assert res.results[1].length == 16
            assert res.results[1].data == list(b"\xaa\x55")

            ms.push_packet(b"\xff\xff\xff\xff\x0d\xf0\xad\xba")
            res = xm.dbgReadModifyWrite(1, 4, 0xCAFEBABE, 0xFFFFFFFF, 0xDEADBEEF)
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x18,
                        0x00,
                        0x0C,
                        0x00,
                        0xC0,
                        0xFC,
                        0x0B,
                        0x00,
                        0x01,
                        0x04,
                        0x00,
                        0x00,
                        0xBE,
                        0xBA,
                        0xFE,
                        0xCA,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0xFF,
                        0xFF,
                        0xFF,
                        0xFF,
                        0xEF,
                        0xBE,
                        0xAD,
                        0xDE,
                    ]
                )
            )

            assert res.value == 0xBAADF00D

            ms.push_packet(b"\xff")
            res = xm.dbgWrite(1, 4, 0xAAAA0001, [0xDDDD0001, 0xDDDD0002])
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x18,
                        0x00,
                        0x0D,
                        0x00,
                        0xC0,
                        0xFC,
                        0x0C,
                        0x00,
                        0x01,
                        0x04,
                        0x02,
                        0x00,
                        0x01,
                        0x00,
                        0xAA,
                        0xAA,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0x01,
                        0x00,
                        0xDD,
                        0xDD,
                        0x02,
                        0x00,
                        0xDD,
                        0xDD,
                    ]
                )
            )

            assert res == b""

            ms.push_packet(b"\xff")
            res = xm.dbgWriteNext(2, [0xDDDD0003, 0xDDDD0004])
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x10,
                        0x00,
                        0x0E,
                        0x00,
                        0xC0,
                        0xFC,
                        0x0D,
                        0x00,
                        0x02,
                        0x00,
                        0x00,
                        0x00,
                        0x03,
                        0x00,
                        0xDD,
                        0xDD,
                        0x04,
                        0x00,
                        0xDD,
                        0xDD,
                    ]
                )
            )

            assert res == b""

            ms.push_packet(b"\xff")
            res = xm.dbgWriteCan1(1, 0xAAAA0001)
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x0F,
                        0x00,
                        0xC0,
                        0xFC,
                        0x0E,
                        0x01,
                        0x01,
                        0x00,
                        0xAA,
                        0xAA,
                    ]
                )
            )

            assert res == b""

            ms.push_packet(b"\xff")
            res = xm.dbgWriteCan2(4, 2)
            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x10, 0x00, 0xC0, 0xFC, 0x0F, 0x04, 0x02]))

            assert res == b""

            ms.push_packet(b"\xff")
            res = xm.dbgWriteCanNext(1, [0xCAFEBABE])
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x11,
                        0x00,
                        0xC0,
                        0xFC,
                        0x10,
                        0x01,
                        0xBE,
                        0xBA,
                        0xFE,
                        0xCA,
                    ]
                )
            )

            assert res == b""

            ms.push_packet(b"\xff\xff\xff\xff\x01\x00\xdd\xdd\x02\x00\xdd\xdd")
            res = xm.dbgRead(1, 4, 2, 0xAAAA0001)
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x10,
                        0x00,
                        0x12,
                        0x00,
                        0xC0,
                        0xFC,
                        0x11,
                        0x00,
                        0x01,
                        0x04,
                        0x02,
                        0x00,
                        0x01,
                        0x00,
                        0xAA,
                        0xAA,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                    ]
                )
            )

            assert res.data[0] == 0xDDDD0001
            assert res.data[1] == 0xDDDD0002

            ms.push_packet(b"\xff")
            res = xm.dbgReadCan1(1, 0xAAAA0001)
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x13,
                        0x00,
                        0xC0,
                        0xFC,
                        0x12,
                        0x01,
                        0x01,
                        0x00,
                        0xAA,
                        0xAA,
                    ]
                )
            )

            assert res == b""

            ms.push_packet(b"\xff")
            res = xm.dbgReadCan2(4, 2)
            ms._mock_send.assert_called_with(bytes([0x05, 0x00, 0x14, 0x00, 0xC0, 0xFC, 0x13, 0x04, 0x02]))

            assert res == b""

            ms.push_packet(
                "FF 01 00 00 D0 00 00 00 07 00 00 00 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 05 00"
                " 06 00 00 00 00 00 01 01 00 00 00 00 00 00 00 00 00 00 00 00 00 00 05 00 07 00 00 00 00 00 02 01 00 00"
                "00 00 00 00 00 00 01 00 00 00 00 00 05 00 01 00 00 00 00 00 03 01 00 00 00 00 00 00 00 00 02 00 00 00 "
                "00 00 05 00 04 00 00 00 00 00 04 02 00 00 00 00 00 00 00 00 02 00 00 00 00 00 05 00 04 00 00 00 00 00 "
                "00 00 01 00 00 00 00 00 05 00 05 00 00 00 00 00 05 02 00 00 00 00 00 00 00 00 02 00 00 00 00 00 05 00 "
                "04 00 00 00 00 00 00 00 02 00 00 00 00 00 05 00 05 00 00 00 00 00 06 01 00 00 00 00 00 00 00 00 04 00 "
                "00 00 00 00 05 00 05 00 00 00 00 00"
            )
            res = xm.dbgGetTriDescTbl()
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x08,
                        0x00,
                        0x15,
                        0x00,
                        0xC0,
                        0xFC,
                        0x14,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                    ]
                )
            )

            assert res.mode == 1
            assert res.length == 208
            assert res.table.tri_cnt == 7
            assert res.table.tris[0].tri == 0
            assert res.table.tris[0].trad_cnt == 1
            assert res.table.tris[0].trads[0].trai == 0x0
            assert res.table.tris[0].trads[0].trdt == 0x5
            assert res.table.tris[0].trads[0].trat == 0x6
            assert res.table.tris[1].tri == 1
            assert res.table.tris[1].trad_cnt == 1
            assert res.table.tris[1].trads[0].trai == 0x0
            assert res.table.tris[1].trads[0].trdt == 0x5
            assert res.table.tris[1].trads[0].trat == 0x7
            assert res.table.tris[2].tri == 2
            assert res.table.tris[2].trad_cnt == 1
            assert res.table.tris[2].trads[0].trai == 0x10000
            assert res.table.tris[2].trads[0].trdt == 0x5
            assert res.table.tris[2].trads[0].trat == 0x1
            assert res.table.tris[3].tri == 3
            assert res.table.tris[3].trad_cnt == 1
            assert res.table.tris[3].trads[0].trai == 0x20000
            assert res.table.tris[3].trads[0].trdt == 0x5
            assert res.table.tris[3].trads[0].trat == 0x4
            assert res.table.tris[4].tri == 4
            assert res.table.tris[4].trad_cnt == 2
            assert res.table.tris[4].trads[0].trai == 0x20000
            assert res.table.tris[4].trads[0].trdt == 0x5
            assert res.table.tris[4].trads[0].trat == 0x4
            assert res.table.tris[4].trads[1].trai == 0x10000
            assert res.table.tris[4].trads[1].trdt == 0x5
            assert res.table.tris[4].trads[1].trat == 0x5
            assert res.table.tris[5].tri == 5
            assert res.table.tris[5].trad_cnt == 2
            assert res.table.tris[5].trads[0].trai == 0x20000
            assert res.table.tris[5].trads[0].trdt == 0x5
            assert res.table.tris[5].trads[0].trat == 0x4
            assert res.table.tris[5].trads[1].trai == 0x20000
            assert res.table.tris[5].trads[1].trdt == 0x5
            assert res.table.tris[5].trads[1].trat == 0x5
            assert res.table.tris[6].tri == 6
            assert res.table.tris[6].trad_cnt == 1
            assert res.table.tris[6].trads[0].trai == 0x40000
            assert res.table.tris[6].trads[0].trdt == 0x5
            assert res.table.tris[6].trads[0].trat == 0x5

            ms.push_packet("FF 00 0F 00 81 00 0A 34 00 00 00 00 00 00 00 00 00 C2 03")
            res = xm.dbgLlbt([0x01, 0x00, 0x05, 0x34, 0x00, 0x26, 0x25, 0xA0, 0xDC, 0x03])
            ms._mock_send.assert_called_with(
                bytes(
                    [
                        0x10,
                        0x00,
                        0x16,
                        0x00,
                        0xC0,
                        0xFC,
                        0x15,
                        0x00,
                        0x0A,
                        0x00,
                        0x01,
                        0x00,
                        0x05,
                        0x34,
                        0x00,
                        0x26,
                        0x25,
                        0xA0,
                        0xDC,
                        0x03,
                    ]
                )
            )

            assert res.length == 15
            assert res.data == list(b"\x81\x00\x0a\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc2\x03")

    @mock.patch("pyxcp.transport.eth.socket.socket")
    @mock.patch("pyxcp.transport.eth.selectors.DefaultSelector")
    def testTimeCorrelationProperties(self, mock_selector, mock_socket):
        ms = MockSocket()

        mock_socket.return_value = ms
        mock_selector.return_value = ms

        with Master("eth", config=create_config()) as xm:
            ms.push_packet(self.DefaultConnectResponse)

            res = xm.connect()

            ms._mock_send.assert_called_with(self.DefaultConnectCmd)

            ms.push_packet("FF 15 25 01 1F 00 78 56")

            res = xm.timeCorrelationProperties(0x15, 0x01, 0x1234)

            ms._mock_send.assert_called_with(bytes([0x06, 0x00, 0x01, 0x00, 0xC6, 0x15, 0x01, 0x00, 0x34, 0x12]))

            assert res.slaveConfig == 0x15
            assert res.observableClocks == 0x25
            assert res.syncState == 0x01
            assert res.clockInfo == 0x1F
            assert res.clusterId == 0x5678



#!/usr/bin/env python
"""Implements error-handling according to XCP spec."""

import functools
import logging
import threading
import time
import types
from collections import namedtuple
from typing import Generic, List, Optional, TypeVar

import can

from pyxcp.errormatrix import ERROR_MATRIX, Action, PreAction
from pyxcp.types import COMMAND_CATEGORIES, XcpError, XcpResponseError, XcpTimeoutError


handle_errors = True  # enable/disable XCP error-handling.

# Thread-local flag to suppress logging for expected XCP negative responses


_thread_flags = threading.local()


def set_suppress_xcp_error_log(value: bool) -> None:
    try:
        _thread_flags.suppress_xcp_error_log = bool(value)
    except Exception:
        pass


def is_suppress_xcp_error_log() -> bool:
    try:
        return bool(getattr(_thread_flags, "suppress_xcp_error_log", False))
    except Exception:
        return False


class SingletonBase:
    _lock = threading.Lock()

    def __new__(cls, *args, **kws):
        # Double-Checked Locking
        if not hasattr(cls, "_instance"):
            try:
                cls._lock.acquire()
                if not hasattr(cls, "_instance"):
                    cls._instance = super().__new__(cls)
            finally:
                cls._lock.release()
        return cls._instance


Function = namedtuple("Function", "fun arguments")  # store: var | load: var


class InternalError(Exception):
    """Indicates an internal error, like invalid service."""


class SystemExit(Exception):
    """"""

    def __init__(self, msg: str, error_code: int = None, *args, **kws):
        super().__init__(*args, **kws)
        self.error_code = error_code
        self.msg = msg

    def __str__(self):
        return f"SystemExit(error_code={self.error_code}, message={self.msg!r})"

    __repr__ = __str__


class UnrecoverableError(Exception):
    """"""


def func_name(func):
    return func.__qualname__ if func is not None else None


def getErrorHandler(service):
    """"""
    return ERROR_MATRIX.get(service)


def getTimeoutHandler(service):
    """"""
    handler = getErrorHandler(service)
    if handler is None:
        raise InternalError("Invalid Service")
    return handler.get(XcpError.ERR_TIMEOUT)


def getActions(service, error_code):
    """"""
    error_str = str(error_code)
    if error_code == XcpError.ERR_TIMEOUT:
        preActions, actions = getTimeoutHandler(service)
    else:
        eh = getErrorHandler(service)
        if eh is None:
            raise InternalError(f"Invalid Service 0x{service:02x}")
        # print(f"Try to handle error -- Service: {service.name} Error-Code: {error_code}")
        handler = eh.get(error_str)
        if handler is None:
            raise SystemExit(f"Service {service.name!r} has no handler for {error_code}.", error_code=error_code)
        preActions, actions = handler
    return preActions, actions


def actionIter(actions):
    """Iterate over action from :file:`errormatrix.py`"""
    if isinstance(actions, (tuple, list)):
        yield from actions
    else:
        yield actions


class Arguments:
    """Container for positional and keyword arguments.

    Parameters
    ----------
        args: tuple
            Positional arguments
        kwargs: dict
            Keyword arguments.
    """

    def __init__(self, args=None, kwargs=None):
        if args is None:
            self.args = ()
        else:
            if not hasattr(args, "__iter__"):
                self.args = (args,)
            else:
                self.args = tuple(args)
        self.kwargs = kwargs or {}

    def __str__(self) -> str:
        res = f"{self.__class__.__name__}(ARGS = {self.args}, KWS = {self.kwargs})"
        return res

    def __eq__(self, other) -> bool:
        return (self.args == other.args if other is not None else False) and (
            self.kwargs == other.kwargs if other is not None else False
        )

    __repr__ = __str__


class Repeater:
    """A required action of some XCP errorhandler is repetition.

    Parameters
    ----------
        initial_value: int
            The actual values are predetermined by XCP:
                - REPEAT (one time)
                - REPEAT_2_TIMES (two times)
                - REPEAT_INF_TIMES ("forever")
    """

    REPEAT = 1
    REPEAT_2_TIMES = 2
    INFINITE = -1

    def __init__(self, initial_value: int):
        self._counter = initial_value
        # print("\tREPEATER ctor", hex(id(self)))

    def repeat(self):
        """Check if repetition is required.

        Returns
        -------
            bool
        """
        # print("\t\tCOUNTER:", hex(id(self)), self._counter)
        if self._counter == Repeater.INFINITE:
            return True
        elif self._counter > 0:
            self._counter -= 1
            return True
        else:
            return False


def display_error():
    """Display error information.

    TODO: callback.

    """


class Handler:
    """"""

    def __init__(self, instance, func, arguments, error_code=None):
        self.instance = instance
        if hasattr(func, "__closure__") and func.__closure__:
            self.func = func.__closure__[0].cell_contents  # Use original, undecorated function to prevent
            # nasty recursion problems.
        else:
            self.func = func
        self.arguments = arguments
        self.service = self.instance.service
        self._error_code: int = 0
        if error_code is not None:
            self._error_code = error_code
        self._repeater = None
        self.logger = logging.getLogger("PyXCP")

    def _diagnostics_enabled(self) -> bool:
        try:
            app = getattr(self.instance, "config", None)
            if app is None:
                return True
            general = getattr(app, "general", None)
            if general is None:
                return True
            return bool(getattr(general, "diagnostics_on_failure", True))
        except Exception:
            return True

    def _build_transport_diagnostics(self) -> str:
        try:
            transport = getattr(self.instance, "transport", None)
            if transport is None:
                return ""
            if hasattr(transport, "_build_diagnostics_dump"):
                return transport._build_diagnostics_dump()  # type: ignore[attr-defined]
        except Exception:
            pass
        return ""

    def _append_diag(self, msg: str) -> str:
        # Suppress diagnostics entirely when XCP error logging is suppressed (e.g., try_command probing)
        if is_suppress_xcp_error_log():
            return msg
        if not self._diagnostics_enabled():
            return msg
        diag = self._build_transport_diagnostics()
        if not diag:
            return msg
        # Prefer a Rich-formatted table for compact, readable diagnostics.
        try:
            header = "--- Diagnostics (for troubleshooting) ---"
            body = diag
            if "\n" in diag and diag.startswith("--- Diagnostics"):
                header, body = diag.split("\n", 1)

            # Try to parse the structured JSON body produced by transports
            import json as _json  # Local import to avoid hard dependency at module import time

            payload = _json.loads(body)
            transport_params = payload.get("transport_params") or {}
            last_pdus = payload.get("last_pdus") or []

            # Try to use rich if available
            try:
                from rich.console import Console
                from rich.table import Table
                from rich.panel import Panel
                from textwrap import shorten

                console = Console(file=None, force_terminal=False, width=120, record=True, markup=False)

                # Transport parameters table
                tp_table = Table(title="Transport Parameters", title_style="bold", show_header=True, header_style="bold magenta")
                tp_table.add_column("Key", style="cyan", no_wrap=True)
                tp_table.add_column("Value", style="white")
                from rich.markup import escape as _escape

                for k, v in (transport_params or {}).items():
                    # Convert complex values to compact repr
                    sv = repr(v)
                    sv = shorten(sv, width=80, placeholder="…")
                    tp_table.add_row(_escape(str(k)), _escape(sv))

                # Last PDUs table (most recent last)
                pdu_table = Table(
                    title="Last PDUs (most recent last)", title_style="bold", show_header=True, header_style="bold magenta"
                )
                for col in ("dir", "cat", "ctr", "ts", "len", "data"):
                    pdu_table.add_column(col, no_wrap=(col in {"dir", "cat", "ctr", "len"}), style="white")
                for pdu in last_pdus:
                    try:
                        dir_ = str(pdu.get("dir", ""))
                        cat = str(pdu.get("cat", ""))
                        ctr = str(pdu.get("ctr", ""))
                        # Format timestamp: convert ns -> s with 5 decimals if numeric
                        ts_val = pdu.get("ts", "")
                        try:
                            ts_num = int(ts_val)
                            ts = f"{ts_num / 1_000_000_000:.5f}"
                        except Exception:
                            ts = str(ts_val)
                        ln = str(pdu.get("len", ""))
                        # Prefer showing actual data content; avoid repr quotes
                        data_val = pdu.get("data", "")
                        try:
                            if isinstance(data_val, (bytes, bytearray, list, tuple)):
                                # Lazily import to avoid hard dependency
                                from pyxcp.utils import hexDump as _hexDump

                                data_str = _hexDump(data_val)
                            else:
                                data_str = str(data_val)
                        except Exception:
                            data_str = str(data_val)
                        # Shorten potentially huge values to keep table compact
                        from textwrap import shorten as _shorten

                        ts = _shorten(ts, width=20, placeholder="…")
                        data = _shorten(data_str, width=40, placeholder="…")
                        # Escape strings to avoid Rich markup interpretation (e.g., '[' ']' in hex dumps)
                        dir_e = _escape(dir_)
                        cat_e = _escape(cat)
                        ctr_e = _escape(ctr)
                        ts_e = _escape(ts)
                        ln_e = _escape(ln)
                        data_e = _escape(data)
                        pdu_table.add_row(dir_e, cat_e, ctr_e, ts_e, ln_e, data_e)
                    except Exception:
                        # If anything odd in structure, add a single-cell row with repr
                        from textwrap import shorten as _shorten

                        pdu_table.add_row(_shorten(repr(pdu), width=80, placeholder="…"), "", "", "", "", "")

                # Combine into a single panel and capture as text
                console.print(Panel.fit(tp_table, title=header))
                if last_pdus:
                    console.print(pdu_table)
                rendered = console.export_text(clear=False)

            except Exception:
                # Rich not available or rendering failed; fallback to compact logger lines
                self.logger.error(header)
                if transport_params:
                    self.logger.error("transport_params: %s", transport_params)
                if last_pdus:
                    self.logger.error("last_pdus (most recent last):")
                    for pdu in last_pdus:
                        try:
                            ts_val = pdu.get("ts", "")
                            try:
                                ts_num = int(ts_val)
                                ts_fmt = f"{ts_num / 1_000_000_000:.5f}"
                            except Exception:
                                ts_fmt = str(ts_val)
                            data_val = pdu.get("data", "")
                            if isinstance(data_val, (bytes, bytearray, list, tuple)):
                                from pyxcp.utils import hexDump as _hexDump

                                data_str = _hexDump(data_val)
                            else:
                                data_str = str(data_val)
                            pdu_copy = dict(pdu)
                            pdu_copy["ts"] = ts_fmt
                            pdu_copy["data"] = data_str
                            self.logger.error("%s", pdu_copy)
                        except Exception:
                            self.logger.error("%s", pdu)
        except Exception:
            # As a last resort, emit the whole diagnostics blob verbatim
            try:
                for line in diag.splitlines():
                    self.logger.error(line)
            except Exception:
                pass
        return msg

    def __str__(self):
        return f"Handler(func = {func_name(self.func)} -- {self.arguments} service = {self.service} error_code = {self.error_code})"

    def __eq__(self, other):
        if other is None:
            return False
        return (self.instance == other.instance) and (self.func == other.func) and (self.arguments == other.arguments)

    @property
    def error_code(self) -> int:
        return self._error_code

    @error_code.setter
    def error_code(self, value: int) -> None:
        self._error_code = value

    @property
    def repeater(self):
        # print("\tGet repeater", hex(id(self._repeater)), self._repeater is None)
        return self._repeater

    @repeater.setter
    def repeater(self, value):
        # print("\tSet repeater", hex(id(value)))
        self._repeater = value

    def execute(self):
        self.logger.debug(f"Execute({func_name(self.func)} -- {self.arguments})")
        if isinstance(self.func, types.MethodType):
            return self.func(*self.arguments.args, **self.arguments.kwargs)
        else:
            return self.func(self.instance, *self.arguments.args, **self.arguments.kwargs)

    def actions(self, preActions, actions):
        """Preprocess errorhandling pre-actions and actions."""
        result_pre_actions = []
        result_actions = []
        repetitionCount = 0
        for item in actionIter(preActions):
            if item == PreAction.NONE:
                pass
            elif item == PreAction.WAIT_T7:
                time.sleep(0.02)  # Completely arbitrary for now.
            elif item == PreAction.SYNCH:
                fn = Function(self.instance.synch, Arguments())
                result_pre_actions.append(fn)
            elif item == PreAction.GET_SEED_UNLOCK:
                raise NotImplementedError("Pre-action GET_SEED_UNLOCK")
            elif item == PreAction.SET_MTA:
                fn = Function(self.instance.setMta, Arguments(self.instance.mta))
                result_pre_actions.append(fn)
            elif item == PreAction.SET_DAQ_PTR:
                fn = Function(self.instance.setDaqPtr, Arguments(self.instance.currentDaqPtr))
            elif item == PreAction.START_STOP_X:
                raise NotImplementedError("Pre-action START_STOP_X")
            elif item == PreAction.REINIT_DAQ:
                raise NotImplementedError("Pre-action REINIT_DAQ")
            elif item == PreAction.DISPLAY_ERROR:
                pass
            elif item == PreAction.DOWNLOAD:
                raise NotImplementedError("Pre-action DOWNLOAD")
            elif item == PreAction.PROGRAM:
                raise NotImplementedError("Pre-action PROGRAM")
            elif item == PreAction.UPLOAD:
                raise NotImplementedError("Pre-action UPLOAD")
            elif item == PreAction.UNLOCK_SLAVE:
                resource = COMMAND_CATEGORIES.get(self.instance.service)  # noqa: F841
                raise NotImplementedError("Pre-action UNLOCK_SLAVE")
        for item in actionIter(actions):
            if item == Action.NONE:
                pass
            elif item == Action.DISPLAY_ERROR:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (DISPLAY_ERROR)."), self.error_code)
            elif item == Action.RETRY_SYNTAX:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (RETRY_SYNTAX)."), self.error_code)
            elif item == Action.RETRY_PARAM:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (RETRY_PARAM)."), self.error_code)
            elif item == Action.USE_A2L:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (USE_A2L)."), self.error_code)
            elif item == Action.USE_ALTERATIVE:
                raise SystemExit(
                    self._append_diag("Could not proceed due to unhandled error (USE_ALTERATIVE)."), self.error_code
                )  # TODO: check alternatives.
            elif item == Action.REPEAT:
                repetitionCount = Repeater.REPEAT
            elif item == Action.REPEAT_2_TIMES:
                repetitionCount = Repeater.REPEAT_2_TIMES
            elif item == Action.REPEAT_INF_TIMES:
                repetitionCount = Repeater.INFINITE
            elif item == Action.RESTART_SESSION:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (RESTART_SESSION)."), self.error_code)
            elif item == Action.TERMINATE_SESSION:
                raise SystemExit(
                    self._append_diag("Could not proceed due to unhandled error (TERMINATE_SESSION)."), self.error_code
                )
            elif item == Action.SKIP:
                pass
            elif item == Action.NEW_FLASH_WARE:
                raise SystemExit(self._append_diag("Could not proceed due to unhandled error (NEW_FLASH_WARE)"), self.error_code)
        return result_pre_actions, result_actions, Repeater(repetitionCount)


T = TypeVar("T")


class HandlerStack(Generic[T]):
    """"""

    def __init__(self) -> None:
        self._stack: List[T] = []

    def push(self, value: T):
        if value != self.tos():
            self._stack.append(value)

    def pop(self) -> None:
        if len(self) > 0:
            self._stack.pop()

    def tos(self) -> Optional[T]:
        if len(self) > 0:
            return self._stack[-1]
        else:
            return None
            # raise ValueError("empty stack.")

    def empty(self) -> bool:
        return self._stack == []

    def __len__(self) -> int:
        return len(self._stack)

    def __repr__(self) -> str:
        result = []
        for idx in range(len(self)):
            result.append(str(self[idx]))
        return "\n".join(result)

    def __getitem__(self, ndx: int) -> T:
        return self._stack[ndx]

    __str__ = __repr__


class Executor(SingletonBase):
    """"""

    def __init__(self):
        self.handlerStack = HandlerStack()
        self.repeater = None
        self.logger = logging.getLogger("PyXCP")
        self.previous_error_code = None
        self.error_code = None
        self.func = None
        self.arguments = None

    def __call__(self, inst, func, arguments):
        self.inst = inst
        self.func = func
        self.arguments = arguments
        handler = Handler(inst, func, arguments)
        self.handlerStack.push(handler)
        connect_retries = inst.config.connect_retries
        try:
            while True:
                try:
                    handler = self.handlerStack.tos()
                    res = handler.execute()
                except XcpResponseError as e:
                    # self.logger.critical(f"XcpResponseError [{e.get_error_code()}]")
                    self.error_code = e.get_error_code()
                    handler.error_code = self.error_code
                    try:
                        svc = getattr(inst.service, "name", None)
                        # Derive a human-friendly error name if available
                        try:
                            err_name = (
                                getattr(XcpError, int(self.error_code)).name
                                if hasattr(XcpError, "__members__")
                                else str(self.error_code)
                            )
                        except Exception:
                            # Fallbacks: try enum-style .name or string conversion
                            err_name = getattr(self.error_code, "name", None) or str(self.error_code)
                        try:
                            err_code_int = int(self.error_code)
                        except Exception:
                            err_code_int = self.error_code  # best effort
                        msg = f"XCP negative response: {err_name} (0x{err_code_int:02X})"
                        if svc:
                            msg += f" on service {svc}"
                        # Suppress noisy ERROR log if requested by caller context
                        if is_suppress_xcp_error_log():
                            self.logger.debug(
                                msg,
                                extra={
                                    "event": "xcp_error_suppressed",
                                    "service": svc,
                                    "error_code": err_code_int,
                                    "error_name": err_name,
                                },
                            )
                        else:
                            self.logger.error(
                                msg,
                                extra={
                                    "event": "xcp_error",
                                    "service": svc,
                                    "error_code": err_code_int,
                                    "error_name": err_name,
                                },
                            )
                    except Exception:
                        pass
                except XcpTimeoutError:
                    is_connect = func.__name__ == "connect"
                    self.logger.warning(f"XcpTimeoutError -- Service: {func.__name__!r}")
                    self.error_code = XcpError.ERR_TIMEOUT
                    handler.error_code = self.error_code
                    if is_connect and connect_retries is not None:
                        if connect_retries == 0:
                            raise XcpTimeoutError("Maximum CONNECT retries reached.")
                        connect_retries -= 1
                except TimeoutError:
                    raise
                except can.CanError:
                    # self.logger.critical(f"Exception raised by Python CAN [{str(e)}]")
                    raise
                except Exception:
                    # self.logger.critical(f"Exception [{str(e)}]")
                    raise
                else:
                    self.error_code = None
                    self.handlerStack.pop()
                    if self.handlerStack.empty():
                        return res

                if self.error_code == XcpError.ERR_CMD_SYNCH:
                    # Don't care about SYNCH for now...
                    self.inst.logger.info("SYNCH received.")
                    continue

                if self.error_code is not None:
                    preActions, actions, repeater = handler.actions(*getActions(inst.service, self.error_code))
                    if handler.repeater is None:
                        handler.repeater = repeater
                    for f, a in reversed(preActions):
                        self.handlerStack.push(Handler(inst, f, a, self.error_code))
                self.previous_error_code = self.error_code
                if handler.repeater:
                    if handler.repeater.repeat():
                        continue
                    else:
                        msg = f"Max. repetition count reached while trying to execute service {handler.func.__name__!r}."
                        # Try to append diagnostics from the transport
                        try:
                            if hasattr(handler, "_append_diag"):
                                msg = handler._append_diag(msg)
                        except Exception:
                            pass
                        try:
                            self.logger.error(
                                "XCP unrecoverable",
                                extra={"event": "xcp_unrecoverable", "service": getattr(inst.service, "name", None)},
                            )
                        except Exception:
                            pass
                        raise UnrecoverableError(msg)
        finally:
            # cleanup of class variables
            self.previous_error_code = None
            while not self.handlerStack.empty():
                self.handlerStack.pop()
            self.error_code = None
            self.func = None
            self.arguments = None


def disable_error_handling(value: bool):
    """Disable XCP error-handling (mainly for performance reasons)."""

    global handle_errors
    handle_errors = not bool(value)


def wrapped(func):
    """This decorator is XCP error-handling enabled."""

    @functools.wraps(func)
    def inner(*args, **kwargs):
        if handle_errors:
            inst = args[0]  # First parameter is 'self'.
            arguments = Arguments(args[1:], kwargs)
            executor = Executor()
            res = executor(inst, func, arguments)
        else:
            res = func(*args, **kwargs)
        return res

    return inner



#!/usr/bin/env python

from pyxcp.daq_stim import DaqList


DAQ_INFO = {
    "channels": [
        {
            "cycle": 0,
            "maxDaqList": 1,
            "name": "Key T",
            "priority": 0,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": True, "packed": False, "stim": False},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
        {
            "cycle": 10,
            "maxDaqList": 1,
            "name": "10 ms",
            "priority": 1,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": True, "packed": False, "stim": True},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
        {
            "cycle": 100,
            "maxDaqList": 1,
            "name": "100ms",
            "priority": 2,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": True, "packed": False, "stim": True},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
        {
            "cycle": 1,
            "maxDaqList": 1,
            "name": "1ms",
            "priority": 3,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": True, "packed": False, "stim": True},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
        {
            "cycle": 10,
            "maxDaqList": 1,
            "name": "FilterBypassDaq",
            "priority": 4,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": True, "packed": False, "stim": True},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
        {
            "cycle": 10,
            "maxDaqList": 1,
            "name": "FilterBypassStim",
            "priority": 5,
            "properties": {"consistency": "CONSISTENCY_ODT", "daq": False, "packed": False, "stim": True},
            "unit": "EVENT_CHANNEL_TIME_UNIT_1MS",
        },
    ],
    "processor": {
        "keyByte": {
            "addressExtension": "AE_DIFFERENT_WITHIN_ODT",
            "identificationField": "IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_BYTE",
            "optimisationType": "OM_DEFAULT",
        },
        "maxDaq": 0,
        "minDaq": 0,
        "properties": {
            "bitStimSupported": False,
            "configType": "DYNAMIC",
            "overloadEvent": False,
            "overloadMsb": True,
            "pidOffSupported": False,
            "prescalerSupported": True,
            "resumeSupported": True,
            "timestampSupported": True,
        },
    },
    "resolution": {
        "granularityOdtEntrySizeDaq": 1,
        "granularityOdtEntrySizeStim": 1,
        "maxOdtEntrySizeDaq": 218,
        "maxOdtEntrySizeStim": 218,
        "timestampMode": {"fixed": False, "size": "S4", "unit": "DAQ_TIMESTAMP_UNIT_10US"},
        "timestampTicks": 10,
    },
}

SLAVE_INFO = {
    "addressGranularity": 0,
    "byteOrder": 0,
    "interleavedMode": False,
    "masterBlockMode": True,
    "maxBs": 2,
    "maxCto": 255,
    "maxDto": 1500,
    "maxWriteDaqMultipleElements": 31,
    "minSt": 0,
    "optionalCommMode": True,
    "pgmProcessor": {},
    "protocolLayerVersion": 1,
    "queueSize": 0,
    "slaveBlockMode": True,
    "supportsCalpag": True,
    "supportsDaq": True,
    "supportsPgm": True,
    "supportsStim": True,
    "transportLayerVersion": 1,
    "xcpDriverVersionNumber": 25,
}


class AttrDict(dict):
    def __getattr__(self, name):
        return self[name]


class MockMaster:
    def __init__(self):
        self.slaveProperties = AttrDict(
            {
                "maxDto": 1500,
                "supportsDaq": True,
            }
        )

    def getDaqInfo(self):
        return DAQ_INFO

    def freeDaq(self):
        pass

    def allocDaq(self, daq_count):
        self.daq_count = daq_count

    def allocOdt(self, daq_num, odt_count):
        pass

    def allocOdtEntry(self, daq_num, odt_num, entry_count):
        pass

    def setDaqPtr(self, daqListNumber, odtNumber, odtEntryNumber):
        pass

    def writeDaq(self, bitOffset, entrySize, addressExt, address):
        pass

    def setDaqListMode(self, mode, daqListNumber, eventChannelNumber, prescaler, priority):
        pass

    def startStopDaqList(self, mode, daqListNumber):
        pass

    def startStopSynch(self, mode):
        pass


DAQ_LISTS = [
    DaqList(
        "list1",
        1,
        False,
        True,
        [
            ("channel1", 0x1BD004, 0, "U32"),
            ("channel2", 0x1BD008, 0, "U32"),
            ("PWMFiltered", 0x1BDDE2, 0, "U8"),
            ("PWM", 0x1BDDDF, 0, "U8"),
            ("Triangle", 0x1BDDDE, 0, "U8"),
        ],
    ),
    DaqList(
        "list2",
        3,
        False,
        True,
        [
            ("TestWord_002", 0x1BE124, 0, "U16"),
            ("TestWord_003", 0x1BE128, 0, "U16"),
            ("TestWord_001", 0x1BE120, 0, "U16"),
            ("TestWord_003", 0x1BE128, 0, "U16"),
            ("TestWord_004", 0x1BE12C, 0, "U16"),
            ("TestWord_005", 0x1BE134, 0, "U16"),
            ("TestWord_006", 0x1BE134, 0, "U16"),
            ("TestWord_007", 0x1BE138, 0, "U16"),
            ("TestWord_008", 0x1BE13C, 0, "U16"),
            ("TestWord_009", 0x1BE140, 0, "U16"),
            ("TestWord_011", 0x1BE148, 0, "U16"),
        ],
    ),
]

# daq = DaqProcessor(DAQ_LISTS)
# daq.set_master(MockMaster())
# daq.setup()
# daq.start()



#!/usr/bin/env python
import abc
import logging
import threading
from collections import deque
from typing import Any, Dict, Optional, Type
from pyxcp.timing import Timing
import pyxcp.types as types

from pyxcp.cpp_ext.cpp_ext import Timestamp, TimestampType
from pyxcp.transport.transport_ext import (
    FrameCategory,
    FrameAcquisitionPolicy,
    LegacyFrameAcquisitionPolicy,
    XcpFraming,
    XcpFramingConfig,
    XcpTransportLayerType,  # noqa: F401
    ChecksumType,  # noqa: F401
)
from pyxcp.utils import (
    CurrentDatetime,
    hexDump,
    seconds_to_nanoseconds,
    short_sleep,
)


class EmptyFrameError(Exception):
    """Raised when an empty frame is received."""


def parse_header_format(header_format: str) -> tuple:
    """SxI and USB framing is configurable."""
    if header_format == "HEADER_LEN_BYTE":
        return 1, 0, 0
    elif header_format == "HEADER_LEN_CTR_BYTE":
        return 1, 1, 0
    elif header_format == "HEADER_LEN_FILL_BYTE":
        return 1, 0, 1
    elif header_format == "HEADER_LEN_WORD":
        return 2, 0, 0
    elif header_format == "HEADER_LEN_CTR_WORD":
        return 2, 2, 0
    elif header_format == "HEADER_LEN_FILL_WORD":
        return 2, 0, 2
    else:
        raise ValueError(f"Invalid header format: {header_format}")


class BaseTransport(metaclass=abc.ABCMeta):
    """Base class for transport-layers (Can, Eth, Sxi).

    Parameters
    ----------
    config: dict-like
        Parameters like bitrate.
    loglevel: ["INFO", "WARN", "DEBUG", "ERROR", "CRITICAL"]
        Controls the verbosity of log messages.

    """

    def __init__(
        self,
        config,
        framing_config: XcpFramingConfig,
        policy: Optional[FrameAcquisitionPolicy] = None,
        transport_layer_interface: Optional[Any] = None,
    ):
        self.has_user_supplied_interface: bool = transport_layer_interface is not None
        self.transport_layer_interface: Optional[Any] = transport_layer_interface
        self.parent = None
        self.framing = XcpFraming(framing_config)
        self.policy: FrameAcquisitionPolicy = policy or LegacyFrameAcquisitionPolicy()
        self.closeEvent: threading.Event = threading.Event()

        self.command_lock: threading.Lock = threading.Lock()
        self.policy_lock: threading.Lock = threading.Lock()

        self.logger = logging.getLogger("PyXCP")
        self._debug: bool = self.logger.level == 10
        if transport_layer_interface:
            self.logger.info(f"Transport - User Supplied Transport-Layer Interface: '{transport_layer_interface!s}'")
        self.counter_received: int = -1
        self.create_daq_timestamps: bool = config.create_daq_timestamps
        self.timestamp = Timestamp(TimestampType.ABSOLUTE_TS)
        self._start_datetime: CurrentDatetime = CurrentDatetime(self.timestamp.initial_value)
        self.alignment: int = config.alignment
        self.timeout: int = seconds_to_nanoseconds(config.timeout)
        self.timer_restart_event: threading.Event = threading.Event()
        self.timing: Timing = Timing()
        self.resQueue: deque = deque()
        self.listener: threading.Thread = threading.Thread(
            target=self.listen,
            args=(),
            kwargs={},
            daemon=True,
        )

        self.first_daq_timestamp: Optional[int] = None
        # self.timestamp_origin = self.timestamp.value
        # self.datetime_origin = datetime.fromtimestamp(self.timestamp_origin)
        self.pre_send_timestamp: int = self.timestamp.value
        self.post_send_timestamp: int = self.timestamp.value
        self.recv_timestamp: int = self.timestamp.value
        # Ring buffer for last PDUs to aid diagnostics on failures
        try:
            from collections import deque as _dq

            self._last_pdus = _dq(maxlen=200)
        except Exception:
            self._last_pdus = []

    def __del__(self) -> None:
        self.finish_listener()
        self.close_connection()

    def load_config(self, config) -> None:
        """Load configuration data."""
        class_name: str = self.__class__.__name__.lower()
        self.config: Any = getattr(config, class_name)

    def close(self) -> None:
        """Close the transport-layer connection and event-loop."""
        self.finish_listener()
        # Avoid indefinite blocking on buggy threads
        try:
            if self.listener.is_alive():
                self.listener.join(timeout=2.0)
        except Exception:
            pass
        self.close_connection()

    @abc.abstractmethod
    def connect(self) -> None:
        pass

    def get(self):
        """Get an item from a deque considering a timeout condition."""
        start: int = self.timestamp.value
        while not self.resQueue:
            if self.timer_restart_event.is_set():
                start: int = self.timestamp.value
                self.timer_restart_event.clear()
            if self.timestamp.value - start > self.timeout:
                raise EmptyFrameError
            short_sleep()
        item = self.resQueue.popleft()
        return item

    @property
    def start_datetime(self) -> int:
        """datetime of program start.

        Returns
        -------
        int
        """
        return self._start_datetime

    def start_listener(self):
        if self.listener.is_alive():
            self.finish_listener()
            # Avoid indefinite blocking on buggy threads
            self.listener.join(timeout=2.0)

        # Ensure the close event is cleared before starting a new listener thread.
        if hasattr(self, "closeEvent"):
            self.closeEvent.clear()

        self.listener = threading.Thread(target=self.listen, daemon=True)
        self.listener.start()

    def finish_listener(self):
        if hasattr(self, "closeEvent"):
            self.closeEvent.set()

    def _request_internal(self, cmd, ignore_timeout=False, *data):
        with self.command_lock:
            frame = self._prepare_request(cmd, *data)
            self.timing.start()
            with self.policy_lock:
                self.policy.feed(FrameCategory.CMD, self.framing.counter_send, self.timestamp.value, frame)
            self.send(frame)
            try:
                xcpPDU = self.get()
            except EmptyFrameError:
                if not ignore_timeout:
                    MSG = f"Response timed out (timeout={self.timeout / 1_000_000_000}s)"
                    with self.policy_lock:
                        self.policy.feed(
                            FrameCategory.METADATA, self.framing.counter_send, self.timestamp.value, bytes(MSG, "ascii")
                        ) if self._diagnostics_enabled() else ""
                    self.logger.debug("XCP request timeout", extra={"event": "timeout", "command": cmd.name})
                    raise types.XcpTimeoutError(MSG) from None
                else:
                    self.timing.stop()
                    return
            self.timing.stop()
            pid = types.Response.parse(xcpPDU).type
            if pid == "ERR" and cmd.name != "SYNCH":
                with self.policy_lock:
                    self.policy.feed(FrameCategory.ERROR, self.counter_received, self.timestamp.value, xcpPDU[1:])
                err = types.XcpError.parse(xcpPDU[1:])
                raise types.XcpResponseError(err)
            return xcpPDU[1:]

    def request(self, cmd, *data):
        return self._request_internal(cmd, False, *data)

    def request_optional_response(self, cmd, *data):
        return self._request_internal(cmd, True, *data)

    def block_request(self, cmd, *data):
        """
        Implements packet transmission for block communication model (e.g. DOWNLOAD block mode)
        All parameters are the same as in request(), but it does not receive response.
        """

        # check response queue before each block request, so that if the slave device
        # has responded with a negative response (e.g. ACCESS_DENIED or SEQUENCE_ERROR), we can
        # process it.
        if self.resQueue:
            xcpPDU = self.resQueue.popleft()
            pid = types.Response.parse(xcpPDU).type
            if pid == "ERR" and cmd.name != "SYNCH":
                err = types.XcpError.parse(xcpPDU[1:])
                raise types.XcpResponseError(err)
        with self.command_lock:
            if isinstance(data, list):
                data = data[0]  # C++ interfacing.
            frame = self._prepare_request(cmd, *data)
            with self.policy_lock:
                self.policy.feed(
                    FrameCategory.CMD if int(cmd) >= 0xC0 else FrameCategory.STIM,
                    self.framing.counter_send,
                    self.timestamp.value,
                    frame,
                )
            self.send(frame)

    def _prepare_request(self, cmd, *data):
        """
        Prepares a request to be sent
        """
        if self._debug:
            self.logger.debug(cmd.name)
        self.parent._setService(cmd)
        frame = self.framing.prepare_request(cmd, *data)
        if self._debug:
            self.logger.debug(f"-> {hexDump(frame)}")
        return frame

    def block_receive(self, length_required: int) -> bytes:
        """
        Implements packet reception for block communication model
        (e.g. for XCP on CAN)

        Parameters
        ----------
        length_required: int
            number of bytes to be expected in block response packets

        Returns
        -------
        bytes
            all payload bytes received in block response packets

        Raises
        ------
        :class:`pyxcp.types.XcpTimeoutError`
        """
        block_response = b""
        start = self.timestamp.value
        while len(block_response) < length_required:
            if len(self.resQueue):
                partial_response = self.resQueue.popleft()
                block_response += partial_response[1:]
            else:
                if self.timestamp.value - start > self.timeout:
                    waited = (self.timestamp.value - start) / 1e9 if hasattr(self.timestamp, "value") else None
                    msg = f"Response timed out [block_receive]: received {len(block_response)} of {length_required} bytes"
                    if waited is not None:
                        msg += f" after {waited:.3f}s"
                    # Attach diagnostics
                    # diag = self._build_diagnostics_dump() if self._diagnostics_enabled() else ""
                    self.logger.debug("XCP block_receive timeout", extra={"event": "timeout"})
                    raise types.XcpTimeoutError(msg) from None
                short_sleep()
        return block_response

    @abc.abstractmethod
    def send(self, frame):
        pass

    @abc.abstractmethod
    def close_connection(self):
        """Does the actual connection shutdown.
        Needs to be implemented by any sub-class.
        """
        pass

    @abc.abstractmethod
    def listen(self):
        pass

    def process_event_packet(self, packet):
        packet = packet[1:]
        ev_type = packet[0]
        self.logger.debug(f"EVENT-PACKET: {hexDump(packet)}")
        if ev_type == types.Event.EV_CMD_PENDING:
            self.timer_restart_event.set()

    def process_response(self, response: bytes, length: int, counter: int, recv_timestamp: int) -> None:
        # Important: determine PID first so duplicate counter handling can be applied selectively.
        pid = response[0]

        if pid >= 0xFC:
            # Do not drop RESPONSE/EVENT/SERV frames even if the transport counter repeats.
            # Some slaves may reuse the counter while DAQ traffic is active, and we must not lose
            # command responses; otherwise request() can stall until timeout.
            if self._debug:
                self.logger.debug(f"<- L{length} C{counter} {hexDump(response)}")
            self.counter_received = counter
            # Record incoming non-DAQ frames for diagnostics
            self._record_pdu(
                "in",
                (FrameCategory.RESPONSE if pid >= 0xFE else FrameCategory.SERV if pid == 0xFC else FrameCategory.EVENT),
                counter,
                recv_timestamp,
                response,
                length,
            )
            if pid >= 0xFE:
                self.resQueue.append(response)
                with self.policy_lock:
                    self.policy.feed(FrameCategory.RESPONSE, self.counter_received, self.timestamp.value, response)
                self.recv_timestamp = recv_timestamp
            elif pid == 0xFD:
                self.process_event_packet(response)
                with self.policy_lock:
                    self.policy.feed(FrameCategory.EVENT, self.counter_received, self.timestamp.value, response)
            elif pid == 0xFC:
                with self.policy_lock:
                    self.policy.feed(FrameCategory.SERV, self.counter_received, self.timestamp.value, response)
        else:
            # DAQ traffic: Some transports reuse or do not advance the counter for DAQ frames.
            # Do not drop DAQ frames on duplicate counters to avoid losing measurements.
            if counter == self.counter_received:
                self.logger.debug(f"Duplicate message counter {counter} received (DAQ) - not dropping")
                # DAQ still flowing – reset request timeout window to avoid false timeouts while
                # the slave is busy but has not yet responded to a command.
                self.timer_restart_event.set()
                # Fall through and process the frame as usual.
            self.counter_received = counter
            if self._debug:
                self.logger.debug(f"<- L{length} C{counter} ODT_Data[0:8] {hexDump(response[:8])}")
            if self.first_daq_timestamp is None:
                self.first_daq_timestamp = recv_timestamp
            if self.create_daq_timestamps:
                timestamp = recv_timestamp
            else:
                timestamp = 0
            # Record DAQ frame (only keep small prefix in payload string later)
            self._record_pdu("in", FrameCategory.DAQ, counter, timestamp, response, length)
            # DAQ activity indicates the slave is alive/busy; keep extending the wait window for any
            # outstanding request, similar to EV_CMD_PENDING behavior on stacks that don't emit it.
            self.timer_restart_event.set()
            with self.policy_lock:
                self.policy.feed(FrameCategory.DAQ, self.counter_received, timestamp, response)

    def _record_pdu(
        self,
        direction: str,
        category: FrameCategory,
        counter: int,
        timestamp: int,
        payload: bytes,
        length: Optional[int] = None,
    ) -> None:
        try:
            entry = {
                "dir": direction,
                "cat": category.name,
                "ctr": int(counter),
                "ts": int(timestamp),
                "len": int(length if length is not None else len(payload)),
                "data": hexDump(payload if category != FrameCategory.DAQ else payload[:8])[:512],
            }
            self._last_pdus.append(entry)
        except Exception:
            pass  # nosec

    def _build_diagnostics_dump(self) -> str:
        import json as _json

        # transport params
        tp = {"transport": self.__class__.__name__}
        cfg = getattr(self, "config", None)
        # Extract common Eth/Can fields when available
        for key in (
            "host",
            "port",
            "protocol",
            "ipv6",
            "bind_to_address",
            "bind_to_port",
            "fd",
            "bitrate",
            "data_bitrate",
            "can_id_master",
            "can_id_slave",
        ):
            if cfg is not None and hasattr(cfg, key):
                try:
                    tp[key] = getattr(cfg, key)
                except Exception:
                    pass  # nosec
        last_n = 20
        try:
            app = getattr(self.config, "parent", None)
            app = getattr(app, "parent", None)
            if app is not None and hasattr(app, "general") and hasattr(app.general, "diagnostics_last_pdus"):
                last_n = int(app.general.diagnostics_last_pdus or last_n)
        except Exception:
            pass  # nosec
        pdus = list(self._last_pdus)[-last_n:]
        payload = {
            "transport_params": tp,
            "last_pdus": pdus,
        }
        try:
            body = _json.dumps(payload, ensure_ascii=False, default=str, indent=2)
        except Exception:
            body = str(payload)
        # Add a small header to explain what follows
        header = "--- Diagnostics (for troubleshooting) ---"
        return f"{header}\n{body}"

    def _diagnostics_enabled(self) -> bool:
        try:
            app = getattr(self.config, "parent", None)
            app = getattr(app, "parent", None)
            if app is not None and hasattr(app, "general"):
                return bool(getattr(app.general, "diagnostics_on_failure", True))
        except Exception:
            return True
        return True

    # @abc.abstractproperty
    # @property
    # def transport_layer_interface(self) -> Any:
    #    pass

    # @transport_layer_interface.setter
    # def transport_layer_interface(self, value: Any) -> None:
    #    self._transport_layer_interface = value


def create_transport(name: str, *args, **kws) -> BaseTransport:
    """Factory function for transports.

    Returns
    -------
    :class:`BaseTransport` derived instance.
    """
    name = name.lower()
    transports = available_transports()
    if name in transports:
        transport_class: Type[BaseTransport] = transports[name]
    else:
        raise ValueError(f"{name!r} is an invalid transport -- please choose one of [{' | '.join(transports.keys())}].")
    return transport_class(*args, **kws)


def available_transports() -> Dict[str, Type[BaseTransport]]:
    """List all subclasses of :class:`BaseTransport`.

    Returns
    -------
    dict
        name: class
    """
    transports = BaseTransport.__subclasses__()
    return {t.__name__.lower(): t for t in transports}



from time import perf_counter

from pyxcp.recorder import FrameCategory, XcpLogFileReader, XcpLogFileWriter


# Pre-allocate a 100MB file -- Note: due to the nature of memory-mapped files, this is a HARD limit.
# Chunk size is 1MB (i.e. compression granularity).
writer = XcpLogFileWriter("test_logger", prealloc=100, chunk_size=1)

# Write some records.
start_time = perf_counter()
for idx in range(512 * 1024):
    value = idx % 256
    writer.add_frame(FrameCategory.CMD, idx, perf_counter() - start_time, bytes([value] * value))
writer.finalize()  # We're done.


reader = XcpLogFileReader("test_logger")
hdr = reader.get_header()  # Get file information.
print(hdr)

df = reader.as_dataframe()  # Return recordings as Pandas DataFrame.
print(df.info())
print(df.describe())
print(df)

reader.reset_iter()  # Non-standard method to restart iteration.

idx = 0
# Iterate over frames/records.
for _frame in reader:
    idx += 1
print("---")
print(f"Iterated over {idx} frames")



#!/usr/bin/env python
# from .transport_ext import (
#    FrameAcquisitionPolicy,  # noqa: F401
#    FrameRecorderPolicy,  # noqa: F401
#    LegacyFrameAcquisitionPolicy,  # noqa: F401
#    NoOpPolicy,  # noqa: F401
#    StdoutPolicy,  # noqa: F401
# )

from .can import Can  # noqa: F401
from .eth import Eth  # noqa: F401
from .sxi import SxI  # noqa: F401
from .usb_transport import Usb  # noqa: F401



#!/usr/bin/env python
"""Very basic hello-world example."""

import argparse

from pyxcp.cmdline import ArgumentParser


parser = argparse.ArgumentParser(description="XCP unlock example")
parser.add_argument(
    "-s",
    "--sk-dll",
    dest="sk_dll",
    help="Seed-and-Key .DLL name",
    type=str,
    default=None,
)

ap = ArgumentParser(parser)

with ap.run() as x:
    if ap.args.sk_dll:
        x.seedNKeyDLL = ap.args.sk_dll

    x.connect()

    print("")
    rps = x.getCurrentProtectionStatus()
    print("Protection before unlocking:", rps, end="\n\n")

    x.cond_unlock()

    rps = x.getCurrentProtectionStatus()
    print("Protection after unlocking:", rps)

    x.disconnect()



#!/usr/bin/env python
"""
Reusable command-line parser utilities.
"""

from __future__ import annotations

import argparse
import sys
from collections.abc import Sequence


class StrippingParser:
    """Base class for parsers that can strip recognized arguments from sys.argv.

    This is useful when multiple argument parsers are used in a single process (chained),
    and downstream parsers should not see arguments already handled by an
    upstream parser.
    """

    def __init__(self, parser: argparse.ArgumentParser) -> None:
        """Initialize with an existing ArgumentParser instance."""
        self._parser = parser

    @property
    def parser(self) -> argparse.ArgumentParser:
        """Return the underlying ArgumentParser instance."""
        return self._parser

    def parse_known(self, argv: Sequence[str] | None = None) -> tuple[argparse.Namespace, list[str]]:
        """Parse known args from ``argv`` and return (namespace, remaining).

        Does not mutate ``sys.argv``.
        """
        if argv is None:
            argv = sys.argv[1:]
        namespace, remaining = self._parser.parse_known_args(argv)

        # Emulate standard -h/--help behavior if "help" is in the namespace
        if getattr(namespace, "help", False):
            # Print help and exit similarly to ArgumentParser
            # We need a temporary full parser to render standard help including program name
            full = argparse.ArgumentParser(description=self._parser.description)
            for action in self._parser._actions:
                if action.option_strings and action.dest == "help":
                    continue
                full._add_action(action)
            full.print_help()
            sys.exit(0)

        return namespace, list(remaining)

    def strip_from_argv(self, argv: list[str] | None = None) -> None:
        """Remove this parser's recognized options from ``sys.argv`` in-place.

        If ``argv`` is provided, it is treated as a mutable list whose content
        will be replaced with the stripped version (first element preserved as
        program name). If omitted, ``sys.argv`` is modified.
        """
        arg_list = argv if argv is not None else sys.argv
        if not arg_list:
            return
        _, remaining = self.parse_known(arg_list[1:])
        # Rebuild argv with program name + remaining
        prog = arg_list[0]
        arg_list[:] = [prog] + remaining

    def parse_and_strip(self, argv: list[str] | None = None) -> argparse.Namespace:
        """Parse options and strip them from argv; returns the parsed namespace."""
        arg_list = argv if argv is not None else sys.argv
        if not arg_list:
            # Fallback for empty list
            return self.parse_known([])[0]
        namespace, remaining = self.parse_known(arg_list[1:])
        # mutate
        prog = arg_list[0]
        arg_list[:] = [prog] + remaining
        return namespace



#!/usr/bin/env python
"""
Parse (transport-layer specific) command line parameters
and create a XCP master instance.
"""

import argparse
import warnings

from pyxcp.config import (  # noqa: F401
    create_application,
    get_application,
    reset_application,
)
from pyxcp.master import Master
from pyxcp.utils.cli import StrippingParser


warnings.simplefilter("always")


class ArgumentParser:
    """Argument parser for pyXCP applications.

    This class provides a way to add custom command-line arguments to pyXCP applications.
    It also supports a callout function that will be called with the master instance and
    the parsed arguments.
    """

    def __init__(self, user_parser=None, description=None, *args, **kws):
        if isinstance(user_parser, argparse.ArgumentParser):
            self._parser = StrippingParser(user_parser)
            self._callout = None
        else:
            # Create a default parser. user_parser might be a callout function or None.
            parser = argparse.ArgumentParser(description=description)
            self._parser = StrippingParser(parser)
            self._callout = user_parser
        self._description = description
        self.args = self._parser.parse_and_strip()

    def run(self, policy=None, transport_layer_interface=None):
        """Create and configure a synchronous master instance.

        Args:
            policy: Optional policy to use for the master
            transport_layer_interface: Optional transport layer interface to use

        Returns:
            A configured master instance
        """
        # Create the application with custom arguments and callout
        application = get_application(options=[], callout=self._callout)

        # Create the master instance
        master = Master(
            application.transport.layer, config=application, policy=policy, transport_layer_interface=transport_layer_interface
        )

        # If there's a callout function, call it with the master and args
        if application.callout is not None and hasattr(application, "custom_args"):
            args = application.custom_args.get_args()
            application.callout(master, args)

        return master

    @property
    def parser(self):
        return self._parser.parser
